from copy import deepcopy

from Bio.SeqFeature import CompoundLocation
from Bio.SeqFeature import SeqFeature
from Bio.SeqRecord import SeqRecord


def feature_to_json(feature: SeqFeature, length: int) -> dict:
    location = feature.location
    if isinstance(location, CompoundLocation):
        start = location.parts[0].start
        end = location.parts[-1].end
        strand = location.parts[0].strand
        assert location.parts[0].end == length
    else:
        start = int(feature.location.start)
        end = int(feature.location.end)
        strand = feature.location.strand
    qualifier_label = feature.qualifiers["label"]
    if isinstance(qualifier_label, list):
        name = qualifier_label[0]
    elif isinstance(qualifier_label, str):
        name = qualifier_label
    elif qualifier_label is None:
        name = "misc_feature"
    else:
        raise ValueError(
            "Type '{}' {} is not a valid feature name".format(
                type(qualifier_label), qualifier_label
            )
        )
    color = feature.qualifiers["ApEinfo_fwdcolor"][0]
    if end >= length:
        end = 0
    if start >= length:
        start = 0
    return {
        "start": start,
        "end": end,
        "strand": strand,
        "color": color,
        "name": name,
        "type": feature.type,
    }


def seqrecord_to_json(record: SeqRecord) -> dict:
    feature_annotations = [feature_to_json(f, len(record.seq)) for f in record.features]
    data = {
        "bases": str(record.seq),
        "length": len(record.seq),
        "name": record.name,
        "id": record.id,
        "annotations": feature_annotations,
        "customFields": deepcopy(record.annotations),
    }
    annotations = deepcopy(record.annotations)

    if annotations.get("topology", "linear") == "circular":
        annotations["isCircular"] = True
    else:
        annotations["isCircular"] = False
    if "topology" in annotations:
        del annotations["topology"]
    data.update(annotations)
    return data



import bisect
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple

import networkx as nx

from dasi.exceptions import DasiInvalidMolecularAssembly
from dasi.models import AlignmentContainer
from dasi.models import Assembly
from dasi.models import AssemblyNode


class DesignResult(Iterable):
    """DesignResult container.

    Maintains a list of top assemblies.
    """

    def __init__(
        self, container: AlignmentContainer, graph: nx.DiGraph, query_key: str
    ):
        self.container = container
        self.graph = graph
        self.query_key = query_key
        self.query = self.container.seqdb[query_key]
        self._assemblies = []
        self._keys = []

    @property
    def seqdb(self):
        return self.container.seqdb

    @property
    def assemblies(self) -> Tuple[Assembly, ...]:
        """Return a tuple of all assemblies.

        :return: tuple of all assemblies.
        """
        return tuple(self._assemblies)

    def _add_assembly_from_path(self, path: List[AssemblyNode]):
        return Assembly(
            path,
            self.container,
            self.graph,
            self.query_key,
            self.query,
            seqdb=self.seqdb,
            do_raise=False,
        )

    def add_assembly(
        self,
        path: List[AssemblyNode],
        ignore_invalid: bool = False,
        allow_invalid: bool = False,
    ):
        """Add an assembly from a list of nodes.

        :param path: list of nodes
        :return: None
        """

        assembly = self._add_assembly_from_path(path)

        # Validate the assembly
        try:
            assembly.post_validate()
        except DasiInvalidMolecularAssembly as e:
            if ignore_invalid:
                return
            elif not allow_invalid:
                raise e

        cost = assembly.compute_cost()
        n_nodes = len(assembly._nodes)
        k = (cost, n_nodes)
        i = bisect.bisect_left(self._keys, k)
        self._assemblies.insert(i, assembly)
        self._keys.insert(i, k)
        return assembly

    def add_assemblies(
        self,
        paths: List[List[AssemblyNode]],
        ignore_invalid: bool = False,
        allow_invalid: bool = False,
    ):
        """Adds a list of assemblies.

        :param paths: list of list of paths
        :return: None
        """
        for path in paths:
            self.add_assembly(
                path, ignore_invalid=ignore_invalid, allow_invalid=allow_invalid
            )

    #
    #
    #     for a in self.assemblies:
    #         for n1, n2, edata in a.edges():
    #             result = edata["sequence_result"]
    #             mol_type = edata["type_def"]
    #             if mol_type.int_or_ext == "internal":
    #                 pass
    #                 # add a new reaction
    #                 # add template
    #                 # add primers
    #             elif mol_type.use_direct:
    #                 pass
    #                 # add fragment
    #             elif mol_type.synthesize:
    #                 pass
    #                 # add synthesize
    #             else:
    #                 pass
    #                 # raise Exception

    def __iter__(self) -> Generator[Assembly, None, None]:
        """Yield assemblies.

        :yield: assembly
        """
        for assembly in self.assemblies:
            yield assembly

    def __getitem__(self, item: str) -> Assembly:
        return list(self)[item]

    def __str__(self):
        return "<{cls}, query={qname} {qk} nassemblies={n}>".format(
            cls=self.__class__.__name__,
            qname=self.query.name,
            qk=self.query_key,
            n=len(self.assemblies),
        )



from .sequence_complexity import count_misprimings_in_amplicon
from .sequence_complexity import DNAStats
from .sequence_generator import generate_fake_designs
from .sequence_generator import generate_fake_library



r"""
Constants (:mod:`dasi.constants`)
=============================

.. currentmodule:: dasi.constants

This module provides DASi constants.
"""


class Constants:
    """DASi constants."""

    ################
    # INTERNAL TYPES
    ################
    FRAGMENT = (
        "PRE-MADE DNA FRAGMENT"
    )  #: an alignment that is generate from an already existing PCR product or fragment
    PCR_PRODUCT = (
        "PCR_PRODUCT"  #: an alignment that is to be generated from a PCR product
    )
    PCR_PRODUCT_WITH_PRIMERS = (
        "PCR_PRODUCT_WITH_PRIMERS"
    )  #: PCR product that can be produces from existing primers
    PCR_PRODUCT_WITH_LEFT_PRIMER = (
        "PCR_PRODUCT_WITH_LEFT_PRIMER"  #: PCR product with existing left primer
    )
    PCR_PRODUCT_WITH_RIGHT_PRIMER = (
        "PCR_PRODUCT_WITH_RIGHT_PRIMER"  #: PCR product with existing right primer
    )
    PRIMER_EXTENSION_PRODUCT_WITH_PRIMERS = (
        "PRIMER_EXTENSION_PRODUCT_WITH_PRIMERS"
    )  #: a pcr product that uses no template, but extends two primers
    PRIMER_EXTENSION_PRODUCT_WITH_LEFT_PRIMER = (
        "PRIMER_EXTENSION_PRODUCT_WITH_LEFT_PRIMER"
    )  #: a pcr product that uses no template, but extends an existing fwd primer.
    PRIMER_EXTENSION_PRODUCT_WITH_RIGHT_PRIMER = (
        "PRIMER_EXTENSION_PRODUCT_WITH_RIGHT_PRIMER"
    )  #: a pcr product that uses no template, but extends an existing rev primer.
    SHARED_SYNTHESIZED_FRAGMENT = (
        "SHARED_SYNTHESIZED_FRAGMENT"
    )  #: explicit definition of a synthesized DNA fragment
    PLASMID = "PLASMID"  #: the designed plasmid

    ################
    # EXTERNAL TYPES
    ################
    GAP = "GAP"  #: region that represents a gap that must be synthesized
    OVERLAP = "OVERLAP"  #: region that represents overlapping molecules

    #################
    # MOLECULES TYPES
    #################
    SHARED_FRAGMENT = (
        "FRAGMENT_SHARED_WITH_OTHER_QUERIES"
    )  #: A fragment alignment that is shared with other queries for potential reuse
    TEMPLATE = "TEMPLATE"  #: A template alignment group. Not an actual molecule.
    MISSING = "__MISSING"  #: missing region
    PRIMER = "PRIMER"  #: a primer binding alignment

    INF = 10.0 ** 6  #: almost infinity

    ############
    # GROUP TAGS
    ############
    PCR_GROUP_TAG = "PCR"
    SHARE_GROUP_TAG = "SHARE"

    ############
    # ALGORITHM
    ############
    ALGORITHM_DEFAULT = "standard optimization"
    ALGORITHM_LIBRARY = "library optimization"



from typing import List
from typing import Union

import numpy as np
import pylab as plt
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D

from dasi.config import Config
from dasi.utils.sequence import DNAStats

# from dasi.design import Design
# from dasi.design import LibraryDesign

# TODO: Implement assembly report
#       1. Coverage report for all groups (total and per each)
#       2. Sequence complexity report
#       3. Spanning distance report
#       4. Assembly report
#       5. Tables (assembly table) and reaction table
#       6. Quilt plot for assembly
# TODO: Report print outs (Markdown, HTML, PDF)


class Report:
    def __init__(self, design: Union["Design", "LibraryDesign"]):
        self.design = design

    def plot_coverage(self, keys: List = None, show: bool = False):
        if not keys:
            keys = self.design.query_keys
        plots = {}
        for qk in keys:
            container = self.design.containers[qk]
            query = self.design.seqdb[qk]
            fig, axes = self.plot_coverage_of_container(container, query)
            plots[qk] = fig, axes
            if show:
                plt.show()
        return plots

    @staticmethod
    def plot_coverage_of_container(container, query):
        arrs = []
        keys = []
        for gkey, groups in container.groups_by_type().items():
            keys.append(gkey)
            a = np.zeros(len(query))
            for group in groups:
                for s in group.query_region.slices():
                    a[s] += 1
            arrs.append(a)
        data = np.vstack(arrs)

        params = {"legend.fontsize": 8}
        plt.rcParams.update(params)

        fig = plt.figure(figsize=(10, 7.5))
        fig.suptitle(query.name, fontsize=16)

        gs = GridSpec(3, 2, hspace=0.55)
        ax1 = fig.add_subplot(gs[0, :2])
        ax2 = fig.add_subplot(gs[1, :2], sharex=ax1)
        ax3 = fig.add_subplot(gs[2, :2], sharex=ax1)

        ax1.set_title("Total Alignment Coverage")
        ax1.set_ylabel("Coverage")
        ax1.set_xlabel("bp")
        ax1.set_yscale("log")
        ax1.plot(np.sum(data, axis=0))

        ax2.set_title("Alignment Coverage")
        ax2.set_ylabel("Coverage")
        ax2.set_xlabel("bp")
        ax2.set_yscale("log")
        ax2.plot(data.T)
        ax2.legend(keys, loc="center left", ncol=1, bbox_to_anchor=(1.0, 0.5))

        stats = DNAStats(str(query.seq) + str(query.seq) + str(query.seq), 14, 20, 20)
        costs_arr = []
        bp_arr = []
        windows = [100, 500, 1000, 2000]
        for window in windows:
            costs = []
            step = min(windows)
            x = np.arange(len(query) - window, len(query) * 2 + window)
            for i in x[::step]:
                costs.append(stats.cost(i, i + window))

            x = x - len(query)
            y = np.repeat(costs, step)

            delta = -(y.shape[0] - x.shape[0])
            if delta == 0:
                delta = None
            y = y[:delta]
            costs_arr.append(y)
            bp_arr.append(x)

        ax3.axhline(
            y=Config.SequenceScoringConfig.complexity_threshold,
            color="k",
            linestyle="-",
        )
        ax3.set_yscale("log")
        ax3.set_xlabel("bp")
        ax3.set_ylabel("Complexity")
        ax3.set_title("Sequence Complexity".format(window))

        for x, y, l in zip(bp_arr, costs_arr, windows):
            ax3.plot(x, y, label=l)
            ax3.legend(title="window (bp)")

        ax1.set_xlim(0, len(query))

        axes = [ax1, ax2, ax3]

        return fig, axes


class AlignmentPlotter:
    def __init__(self, length, starts, ends, types, names, costs, title=""):
        segments = zip(starts, ends, types, names, costs)
        self.segments = list(segments)
        self.length = length

        self.yspacer = 0.3
        self.border = 1
        self.text_spacer = 0.2
        self.title = title
        self.segment_styles = {
            "fragment": {"color": "black", "linewidth": 3},
            "synthesis": {"color": "red", "linewidth": 3},
            "overhang": {"color": "blue", "linewidth": 1},
        }

    def plot(self):

        Y = len([f for f in self.segments if f[2] != "overhang"])

        fig = plt.figure(figsize=(5, self.yspacer * Y))

        ax = fig.gca()
        ax.set_xlim(-self.border, self.length + self.border)
        ax.set_ylim(-self.border, Y + self.border)

        # set border and axis
        ax.axes.get_xaxis().set_visible(True)
        ax.axes.get_yaxis().set_visible(False)
        ax.set_xlabel("bp")
        ax.set_frame_on(False)
        ax.set_title(self.title)
        y = Y
        for i, seg in enumerate(self.segments):
            x1, x2, segtype, name, cost = seg
            if "overlap" in name or "overlap" in segtype:
                segtype = "overhang"
            elif "synthesis" in name or "synthesis" in segtype:
                segtype = "synthesis"
            else:
                segtype = "fragment"
            cost = "{:.1f}".format(float(cost))
            name = "{} - {}".format(Y - y, name)
            segstyle = self.segment_styles[segtype]
            if x1 < x2:
                line = Line2D([seg[0], seg[1]], [y, y], zorder=1, **segstyle)

                # add cpst
                xtext = (seg[0] + seg[1]) / 2.0
                ax.text(
                    xtext,
                    y + self.text_spacer,
                    str(cost),
                    horizontalalignment="center",
                    verticalalignment="bottom",
                )

                # add label
                ax.text(
                    self.length,
                    y,
                    name,
                    horizontalalignment="left",
                    verticalalignment="center",
                )
                ax.add_line(line)
                y -= 1
            else:
                if segtype == "overhang":
                    if i == len(self.segments) - 1:
                        y2 = Y
                    else:
                        y2 = y + 1
                    line1 = Line2D([x1, x2], [y, y2], zorder=0, **segstyle)
                    line2 = Line2D([x2, x1], [y, y2], zorder=0, **segstyle)

                    xtext = (x1 + x2) / 2.0
                    ax.text(
                        xtext,
                        y - 2 * self.text_spacer,
                        str(cost),
                        horizontalalignment="center",
                        verticalalignment="top",
                    )

                    ax.add_line(line1)
                    ax.add_line(line2)
                else:
                    line1 = Line2D([x1, self.length], [y, y], zorder=1, **segstyle)
                    line2 = Line2D([0, x2], [y, y], zorder=1, **segstyle)

                    # add cost
                    if self.length - x1 > x2:
                        xtext = (self.length - x1) / 2.0
                    else:
                        xtext = x2 / 2.0
                    ax.text(
                        xtext,
                        y + self.text_spacer,
                        str(cost),
                        horizontalalignment="center",
                        verticalalignment="bottom",
                    )

                    # add label
                    ax.text(
                        self.length,
                        y,
                        name,
                        horizontalalignment="left",
                        verticalalignment="center",
                    )

                    ax.add_line(line1)
                    ax.add_line(line2)
        plt.show()



r"""
.. _cost_model:

Cost Model (:mod:`dasi.cost`)
=============================

.. currentmodule:: dasi.cost

This module provide cost calculations for molecular assemblies.

Default cost parameters can be opened as follows:

.. code-block::

    cost_model = SpanCost.open()

    # optionally use custom parameters
    cost_model = SpanCost.open("my_parameters.json")

Take a look at the :ref:`parameters schema <cost_schema>` for how to build the
parameter json.

Utilities
---------

.. autosummary::
    :toctree: generated/

    utils
"""
from dasi.cost.cached_span_cost import cached_span_cost
from dasi.cost.span_cost import open_params
from dasi.cost.span_cost import PrimerCostModel
from dasi.cost.span_cost import SpanCost
from dasi.cost.span_cost import SynthesisCostModel
from dasi.cost.span_cost import validate_params



import bisect
import itertools
from functools import partial
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from uuid import uuid4

import networkx as nx
import numpy as np
from Bio.SeqRecord import SeqRecord
from more_itertools import partition
from pyblast.utils import is_circular

from dasi.config import Config
from dasi.constants import Constants
from dasi.cost import SpanCost
from dasi.log import logger
from dasi.models import AlignmentContainer
from dasi.models import AlignmentGroup
from dasi.models import AssemblyNode
from dasi.models import MoleculeType
from dasi.models import MultiPCRProductAlignmentGroup
from dasi.models import PCRProductAlignmentGroup
from dasi.utils import bisect_between
from dasi.utils import Region
from dasi.utils import sort_with_keys
from dasi.utils.sequence import count_misprimings_in_amplicon
from dasi.utils.sequence import DNAStats
from dasi.utils.sequence.sequence_partitioner import find_by_partitions_for_sequence

SequenceScoringConfig = Config.SequenceScoringConfig


# TODO: whenever the efficiency is adjusted, record this in the notes
def add_edge_note(edata, key, value):
    if not edata["notes"]:
        edata["notes"] = {}
    edata["notes"][key] = value


Edge = Tuple[AssemblyNode, AssemblyNode, Dict]


class AssemblyGraphBuilder:
    """Class that builds an AssemblyGraph from an alignment container."""

    COST_THRESHOLD = Config.ASSEMBLY_COST_THRESHOLD

    def __init__(self, alignment_container: AlignmentContainer, span_cost=None):
        self.container = alignment_container
        if span_cost is None:
            self.span_cost = SpanCost.open()
        else:
            self.span_cost = span_cost
        self.G = None
        self.logger = logger(self)

    def add_node(self, node: AssemblyNode) -> None:
        """Add node to the graph.

        :param node: the assembly node to add
        :return: None
        """
        self.G.add_node(AssemblyNode(*node))

    def create_edge(
        self,
        n1: AssemblyNode,
        n2: AssemblyNode,
        cost: Union[float, None],
        material: Union[float, None],
        time: Union[float, None],
        efficiency: Union[float, None],
        span: int,
        type_def: MoleculeType,
        internal_or_external: str,
        condition: Tuple[bool, bool],
        group: Union[AlignmentGroup, PCRProductAlignmentGroup],
        notes: Dict = None,
        add_to_graph: bool = False,
    ):
        """Add an edge between two assembly nodes.

        :param n1: src node
        :param n2: dest node
        :param name: name of the edge
        :param cost: overall cost of the edge
        :param material: material cost of the edge. Used in path calculations.
        :param time: time cost of the edge. Used in path calculations.
        :param efficiency: efficiency of the edge. Used in path calculations.
        :param span: spanning distance (in bp) of the edge.
        :param atype: alignment type of the edge.
        :param kwargs: additional kwargs for the edge data
        :return:
        """

        assert condition == type_def.design
        assert internal_or_external == type_def.int_or_ext
        edge = (
            n1,
            n2,
            dict(
                cost=cost,
                material=material,
                time=time,
                efficiency=efficiency,
                span=span,
                type_def=type_def,
                group=group,
                notes=notes,
            ),
        )
        if add_to_graph:
            self.G.add_edge(edge[0], edge[1], **edge[2])
        return edge

    def iter_internal_edge_data(
        self, group: Union[AlignmentGroup, PCRProductAlignmentGroup]
    ) -> dict:
        q = group.query_region

        mtype = MoleculeType.types[group.type]
        a_expand, b_expand = mtype.design
        internal_cost = mtype.cost
        internal_efficiency = mtype.efficiency

        if q.cyclic:
            # cyclic
            if q.b < q.a:
                pairs = [(q.a, q.b + q.context_length), (q.a + q.context_length, q.b)]
            else:
                pairs = [(q.a, q.b), (q.a + q.context_length, q.b + q.context_length)]
        else:
            # linear
            pairs = [(q.a, q.b)]

        # nodes = []
        # edges = []
        for a, b in pairs:
            if a is not None:
                anode = (a, a_expand, "A")
                # nodes.append(anode)
                if b is not None:
                    bnode = (b, b_expand, "B")
                    # nodes.append(bnode)
                    yield (
                        anode,
                        bnode,
                        dict(
                            material=internal_cost,
                            cost=internal_cost / internal_efficiency,
                            time=0.1,
                            internal_or_external="internal",
                            span=len(group.query_region),
                            type_def=MoleculeType.types[group.type],
                            efficiency=internal_efficiency,
                            condition=(a_expand, b_expand),
                            group=group,
                        ),
                    )
                    # edges.append(edge)
        # return nodes, edges

    def add_internal_edges(
        self, groups: List[Union[AlignmentGroup, PCRProductAlignmentGroup]]
    ):
        """Add 'internal' edges from a list of AlignmentGroups. This means.

        :param groups:
        :return:
        """
        for g in groups:
            for a, b, ab_data in self.iter_internal_edge_data(g):
                for a_overhang, b_overhang in itertools.product(
                    [True, False], repeat=2
                ):
                    a_node = AssemblyNode(a[0], a[1], a[2], a_overhang)
                    b_node = AssemblyNode(b[0], b[1], b[2], b_overhang)
                    self.create_edge(a_node, b_node, add_to_graph=True, **ab_data)

    @staticmethod
    def make_overlap_iterator(
        a: Iterable[AssemblyNode], b: Iterable[AssemblyNode]
    ) -> Generator[Tuple[AssemblyNode, AssemblyNode], None, None]:
        """Find all nodes that satisfy the below condition: b.

        |--------|
               |-------|
               a

        With the exception that when a.index == b.index, this is not
        considered an overlap, but covered in the `make_gap_iterator`,
        due to using bisect.bisect_right. Overlap is more computationally
        intensive.
        """
        a, akeys = sort_with_keys(a, lambda x: x.index)
        b, bkeys = sort_with_keys(b, lambda x: x.index)
        for _a in a:
            i = bisect.bisect_right(bkeys, _a.index)
            for _b in b[i:]:
                yield _b, _a

    @staticmethod
    def make_gap_iterator(
        a: Iterable[AssemblyNode], b: Iterable[AssemblyNode]
    ) -> Generator[Tuple[AssemblyNode, AssemblyNode], None, None]:
        """Find all nodes that satisfy the below condition: b.

        |--------|              |-------|              a
        """
        a, akeys = sort_with_keys(a, lambda x: x.index)
        b, bkeys = sort_with_keys(b, lambda x: x.index)

        for _a in a:
            i = bisect.bisect_right(bkeys, _a.index)
            for _b in b[:i]:
                yield _b, _a

    @staticmethod
    def make_origin_iterator(
        a: Iterable[AssemblyNode], b: Iterable[AssemblyNode], length: int
    ) -> Generator[Tuple[AssemblyNode, AssemblyNode], None, None]:
        a, akeys = sort_with_keys(a, lambda x: x.index)
        b, bkeys = sort_with_keys(b, lambda x: x.index)

        i = bisect.bisect_right(akeys, length)
        j = bisect.bisect_left(bkeys, length)
        for _a in a[:i]:
            for _b in b[j:]:
                yield _b, _a

    def add_external_edges(self, groups, group_keys, nodes: Iterable[AssemblyNode]):
        if not groups:
            return
        query_region = groups[0].query_region
        length = query_region.context_length
        a_nodes, b_nodes = partition(lambda x: x.type == "B", nodes)

        a_nodes = sorted(a_nodes, key=lambda x: x.index)
        b_nodes = sorted(b_nodes, key=lambda x: x.index)

        a_nodes_gap, a_nodes_overhang = [
            list(x) for x in partition(lambda x: x.overhang, a_nodes)
        ]
        b_nodes_gap, b_nodes_overhang = [
            list(x) for x in partition(lambda x: x.overhang, b_nodes)
        ]

        overlap_iter = self.make_overlap_iterator(a_nodes_overhang, b_nodes_overhang)
        gap_iter = self.make_gap_iterator(a_nodes_gap, b_nodes_gap)
        gap_origin_iter = self.make_origin_iterator(a_nodes_gap, b_nodes_gap, length)
        overlap_origin_iter = self.make_origin_iterator(
            a_nodes_overhang, b_nodes_overhang, length
        )

        for bnode, anode in overlap_iter:
            self.add_overlap_edge(bnode, anode, query_region, group_keys, groups)

        for bnode, anode in gap_iter:
            self.add_gap_edge(bnode, anode, query_region)

        for bnode, anode in overlap_origin_iter:
            self.add_overlap_edge(
                bnode, anode, query_region, group_keys, groups, origin=True
            )
        for bnode, anode in gap_origin_iter:
            self.add_gap_edge(bnode, anode, query_region, origin=True)

    def add_overlap_edge(
        self,
        bnode,
        anode,
        query_region,
        group_keys,
        groups,
        origin=False,
        validate_groups_present: bool = True,
        add_to_graph: bool = True,
    ):
        q = query_region.new(anode.index, bnode.index)
        if len(q) == 0:
            return
        filtered_groups = []
        if validate_groups_present:
            if not origin:
                i, j = bisect_between(group_keys, q.a, q.b)
                filtered_groups = groups[i:j]
            else:
                i = bisect.bisect_left(group_keys, q.a)
                j = bisect.bisect_right(group_keys, q.b)
                filtered_groups = groups[i:] + groups[:j]
        if not validate_groups_present or filtered_groups:
            if not origin:
                span = anode.index - bnode.index
            else:
                span = -len(q)
            if span <= 0:
                condition = (bnode.expandable, anode.expandable)
                return self.create_edge(
                    bnode,
                    anode,
                    cost=None,
                    material=None,
                    time=None,
                    efficiency=None,
                    internal_or_external="external",
                    type_def=MoleculeType.types[Constants.OVERLAP](condition),
                    condition=condition,
                    span=span,
                    group=None,
                    add_to_graph=add_to_graph,
                )

    def add_gap_edge(
        self, bnode, anode, query_region, origin=False, add_to_graph: bool = True
    ):
        """Add a 'gap' edge to the graph.

        :param bnode: ending AssemblyNode from the previous edge
        :param anode: starting AssemblyNode from the next edge
        :param query_region: a query region to copy (indices a,b does not matter)
        :param origin: whether this gap spans the origin or not
        :return:
        """
        try:
            if not origin:
                span = anode.index - bnode.index
            else:
                if bnode.index == query_region.context_length:
                    span = len(query_region.new(0, anode.index))
                else:
                    span = len(query_region.new(bnode.index, anode.index))
        except Exception as e:
            print(
                "An exception was raised while attempting to create a abstract region"
            )
            print("origin == " + str(origin))
            print("bnode == " + str(bnode))
            print("anode == " + str(anode))
            print("query_region == " + str(query_region))
            raise e
        if span >= 0:
            condition = (bnode.expandable, anode.expandable)
            # cost_dict = self._get_cost(span, condition)
            return self.create_edge(
                bnode,
                anode,
                cost=None,
                material=None,
                time=None,
                efficiency=None,
                internal_or_external="external",
                type_def=MoleculeType.types[Constants.GAP](condition),
                span=span,
                condition=condition,
                group=None,
                add_to_graph=add_to_graph,
            )

    # TODO: refactor this. One method to update costs of edges. One method to
    #       add edges. One method to remove high cost edges.
    @staticmethod
    def batch_add_edge_costs(graph, edges, span_cost, cost_threshold: float = None):
        """Add costs to all edges at once.

        :return:
        """
        # add external_edge_costs
        edge_dict = {}
        for n1, n2, edata in edges:
            condition = edata["type_def"].design
            assert isinstance(condition, tuple)
            edge_dict.setdefault(condition, []).append(((n1, n2), edata, edata["span"]))

        edges_to_remove = []
        for condition, info in edge_dict.items():
            edges, edata, spans = zip(*info)
            npdf = span_cost.cost(np.array(spans), condition)
            data = npdf.aggregate(np.vstack)
            cost_i = [i for i, col in enumerate(npdf.columns) if col == "cost"][0]

            for i, (e, edge) in enumerate(zip(edata, edges)):
                if cost_threshold is not None and data[cost_i, i] > cost_threshold:
                    edges_to_remove.append(edge)
                else:
                    _d = {c: data[col_i, i] for col_i, c in enumerate(npdf.columns)}
                    if _d["material"] is None:
                        raise Exception
                    e.update(_d)
        if cost_threshold is not None:
            graph.remove_edges_from(edges_to_remove)

    def _get_cost(self, bp, ext):
        data = self.span_cost(bp, ext).data
        return {k: v[0] for k, v in data.items()}

    def update_costs(self, edges: Optional[List[Edge]] = None):
        if edges is None:
            edges = []
            for n1, n2, edata in self.G.edges(data=True):
                if edata["cost"] is None:
                    edges.append((n1, n2, edata))
                elif edata["type_def"].name == Constants.SHARED_SYNTHESIZED_FRAGMENT:
                    edges.append((n1, n2, edata))
        self.batch_add_edge_costs(self.G, edges, self.span_cost, self.COST_THRESHOLD)

    def build_assembly_graph(self) -> nx.DiGraph:

        self.G = nx.DiGraph(name="Assembly Graph")

        groups, group_keys = sort_with_keys(
            self.container.get_groups_by_types(
                [
                    Constants.PCR_PRODUCT,
                    Constants.PCR_PRODUCT_WITH_RIGHT_PRIMER,
                    Constants.PCR_PRODUCT_WITH_LEFT_PRIMER,
                    Constants.PCR_PRODUCT_WITH_PRIMERS,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_PRIMERS,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_RIGHT_PRIMER,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_LEFT_PRIMER,
                    Constants.FRAGMENT,
                    Constants.SHARED_SYNTHESIZED_FRAGMENT,
                ]
            ),
            key=lambda g: g.query_region.a,
        )

        self.add_internal_edges(groups)
        self.add_external_edges(groups, group_keys, self.G.nodes())
        self.update_costs()

        # TODO: freeze?
        # nx.freeze(self.G)
        return self.G


# TODO: evaluate primer designs, scoring PCR products d(how long does this take?)
# TODO: refactor this class, expose config options


class AssemblyGraphPostProcessor:
    """Pre-processing for assembly graphs. Evaluates:

    1. synthesis complexity and weights corresponding edge
    2. pcr product efficiency
    3. (optional) optimal partitions for synthesis fragments
    """

    # TODO: add post processing config

    def __init__(
        self,
        graph: nx.DiGraph,
        query: SeqRecord,
        span_cost: SpanCost,
        seqdb: Dict[str, SeqRecord],
        container: AlignmentContainer,
        stats_repeat_window: Optional[int] = None,
        stats_window: Optional[int] = None,
        stats_hairpin_window: Optional[int] = None,
        edge_threshold: Optional[float] = None,
        stages: Optional[Tuple[str]] = None,
    ):
        if stats_repeat_window is None:
            stats_repeat_window = SequenceScoringConfig.stats_repeat_window
        if stats_window is None:
            stats_window = SequenceScoringConfig.stats_window
        if stats_hairpin_window is None:
            stats_hairpin_window = SequenceScoringConfig.stats_hairpin_window
        if stages is None:
            stages = SequenceScoringConfig.post_process_stages
        if edge_threshold is None:
            edge_threshold = SequenceScoringConfig.edge_threshold
        self.graph = graph
        self.graph_builder = AssemblyGraphBuilder(container, span_cost=span_cost)
        self.graph_builder.G = graph
        self.query = query
        self.seqdb = seqdb
        query_seq = str(query.seq)
        if is_circular(query):
            query_seq = query_seq + query_seq
        self.stats = DNAStats(
            query_seq,
            repeat_window=stats_repeat_window,
            stats_window=stats_window,
            hairpin_window=stats_hairpin_window,
        )
        self.stats_single = DNAStats(
            str(query.seq),
            repeat_window=stats_repeat_window,
            stats_window=stats_window,
            hairpin_window=stats_hairpin_window,
        )
        self.logged_msgs = []
        # TODO: make a more sophisticated complexity function?
        # TODO: expose this to input parameters
        self.COMPLEXITY_THRESHOLD = SequenceScoringConfig.complexity_threshold
        self.logger = logger(self)
        self.span_cost = span_cost
        self.stages = stages
        self.edge_threshold = edge_threshold

    @staticmethod
    def optimize_partition(
        signatures: np.ndarray, step: int, i: int = None, j: int = None
    ):
        """Optimize partition by minimizing the number of signatures in the
        given array.

        :param signatures: array of signatures
        :param step: step size
        :param i:
        :param j:
        :return:
        """
        d = []

        if i is None:
            i = 0
        if j is None:
            j = signatures.shape[1]

        for x in range(i, j, step):
            m1 = np.empty(signatures.shape[1])
            m2 = m1.copy()
            m1.fill(np.nan)
            m2.fill(np.nan)

            m1[:x] = np.random.uniform(1, 10)
            m2[x:] = np.random.uniform(1, 10)

            d += [m1, m2]
        d = np.vstack(d)
        z = np.tile(d, signatures.shape[0]) * signatures.flatten()

        partition_index = np.repeat(
            np.arange(0, signatures.shape[1], step),
            signatures.shape[0] * signatures.shape[1] * 2,
        )

        a, b, c = np.unique(z, return_counts=True, return_index=True)
        i = b[np.where(c > 1)]
        a, c = np.unique(partition_index[i], return_counts=True)
        if len(c):
            arg = c.argmin()
            return a[arg], c[arg]

    def _edge_to_region(self, n1, n2):
        if n1.index == len(self.query) and is_circular(self.query):
            a = 0
            b = n2.index
        else:
            a = n1.index
            b = n2.index
        return Region(a, b, len(self.query), cyclic=is_circular(self.query))

    @staticmethod
    def _adj_eff(edata, e):
        edata["efficiency"] = e
        if e == 0:
            edata["cost"] = np.inf
        if edata["material"] is None:
            edata["cost"] = np.inf
        else:
            edata["cost"] = edata["material"] / edata["efficiency"]

    def _complexity_to_efficiency(self, edata: dict) -> bool:
        if not edata["complexity"] >= 0.0:
            raise ValueError("Complexity is not defined. {}".format(edata))
        ratio = edata["complexity"] / self.COMPLEXITY_THRESHOLD
        if ratio >= 1.0:
            e = SequenceScoringConfig.not_synthesizable_efficiency / ratio
            self._adj_eff(edata, e)
            return True
        return False

    @staticmethod
    def _is_pcr_product(edata):
        return edata["type_def"].name in [
            Constants.PCR_PRODUCT,
            Constants.PCR_PRODUCT_WITH_LEFT_PRIMER,
            Constants.PCR_PRODUCT_WITH_RIGHT_PRIMER,
            Constants.PCR_PRODUCT_WITH_PRIMERS,
        ]

    def score_long_pcr_products(self, n1, n2, edata):
        if self._is_pcr_product(edata):
            span = edata["span"]
            for a, b, c in SequenceScoringConfig.pcr_length_range_efficiency_multiplier:
                if a <= span < b:
                    self._adj_eff(edata, edata["efficiency"] * c)
                    add_edge_note(edata, "long_pcr_product", True)
                    break

    def _score_misprimings_from_alignment(self, alignment):
        subject_key = alignment.subject_key
        subject = self.seqdb[subject_key]
        i = alignment.subject_region.a
        j = alignment.subject_region.b
        subject_seq = str(subject.seq)

        return count_misprimings_in_amplicon(
            subject_seq,
            i,
            j,
            min_primer_anneal=SequenceScoringConfig.mispriming_min_anneal,
            max_primer_anneal=SequenceScoringConfig.mispriming_max_anneal,
            cyclic=alignment.subject_region.cyclic,
        )

    # TODO: select the best template...
    # TODO: speed up this process
    def score_primer_misprimings(self, n1, n2, edata):
        if self._is_pcr_product(edata):

            group = edata["group"]

            misprime_list = []

            if isinstance(group, MultiPCRProductAlignmentGroup):
                prioritize_function = group.prioritize_groupings
                alignments = group.iter_templates()

            elif isinstance(group, AlignmentGroup):
                prioritize_function = group.prioritize_alignments
                alignments = group.alignments
            else:
                raise TypeError(
                    "Group '{}' not supported by this function.".format(group.__class__)
                )

            arr = []
            for index, alignment in enumerate(alignments):
                if not (
                    "PCR" in alignment.type or alignment.type == Constants.FRAGMENT
                ):
                    raise ValueError(
                        "{} is not a valid type for a template".format(alignment.type)
                    )
                mispriming = self._score_misprimings_from_alignment(alignment)
                arr.append((mispriming, index, alignment))
                if mispriming == 0:
                    break
            arr.sort(key=lambda x: x[0])
            indices = [a[1] for a in arr]
            prioritize_function(indices)
            score = arr[0][0]

            self._adj_eff(
                edata,
                edata["efficiency"] * SequenceScoringConfig.mispriming_penalty ** score,
            )
            add_edge_note(edata, "num_misprimings", score)
            add_edge_note(edata, "n_templates_eval", len(misprime_list))

    def _filter_partition_edges(self, edges: List[Edge]) -> List[Edge]:
        edges_to_partition = []
        for n1, n2, edata in edges:
            min_size = (
                edata["type_def"].min_size
                or MoleculeType.types[Constants.SHARED_SYNTHESIZED_FRAGMENT].min_size
            )
            if (
                edata["span"] > min_size * 2
                and edata["complexity"]
                >= Config.SequenceScoringConfig.complexity_threshold
            ):
                edges_to_partition.append((n1, n2, edata))
        return edges_to_partition

    def partition(self, edges: List[Edge]):
        tracker = self.logger.track(
            "INFO", desc="Partitioning sequences", total=3
        ).enter()
        tracker.update(0, "{} highly complex sequences".format(len(edges)))

        edges_to_partition = self._filter_partition_edges(edges)

        cyclic = is_circular(self.query)
        partitions = find_by_partitions_for_sequence(
            self.stats_single,
            cyclic=cyclic,
            threshold=Config.SequenceScoringConfig.complexity_threshold,
            step_size=Config.SequenceScoringConfig.partition_step_size,
            delta=Config.SequenceScoringConfig.partition_overlap,
        )
        tracker.update(1, "Partition: locations: {}".format(partitions))
        add_gap_edge = partial(self.graph_builder.add_gap_edge, add_to_graph=False)
        add_overlap_edge = partial(
            self.graph_builder.add_overlap_edge,
            add_to_graph=True,
            validate_groups_present=False,
            groups=None,
            group_keys=None,
        )

        new_edges = []
        for n1, n2, edata in edges_to_partition:
            r = Region(n1.index, n2.index, len(self.query.seq), cyclic=cyclic)

            if n1.type == "B" and n2.type == "A":

                for p in partitions:
                    if p in r:
                        # TODO: overlap? find optimal partition for overlap?
                        i4 = p
                        i3 = p + Config.SequenceScoringConfig.partition_overlap

                        n3 = AssemblyNode(i3, False, "BC", overhang=True)
                        n4 = AssemblyNode(i4, False, "CA", overhang=True)
                        e1 = add_gap_edge(n1, n3, r, origin=False)
                        e2 = add_gap_edge(n1, n3, r, origin=True)
                        if e1 is None and e2 is None:
                            continue
                        e3 = add_overlap_edge(n3, n4, r, origin=False)
                        e4 = add_overlap_edge(n3, n4, r, origin=True)
                        if e3 is None and e4 is None:
                            continue
                        e5 = add_gap_edge(n4, n2, r, origin=False)
                        e6 = add_gap_edge(n4, n2, r, origin=True)
                        if e5 is None and e6 is None:
                            continue
                        new_edges += [e1, e2, e3, e4, e5, e6]
        for e in new_edges:
            if e is not None:
                self.graph_builder.G.add_edge(e[0], e[1], **e[2])
        edges = []
        for n1, n2, edata in self.graph_builder.G.edges(data=True):
            if edata["cost"] is None:
                edges.append((n1, n2, edata))
        tracker.update(2, "Partition: Added {} new edges".format(len(edges)))
        self.graph_builder.update_costs(edges)
        self.score_complexity_edges(list(self.graph_builder.G.edges(data=True)))
        tracker.exit()

    def remove_inefficient_edges(self):
        to_remove = []
        for n1, n2, edata in self.graph_builder.G.edges(data=True):
            if edata["efficiency"] < self.edge_threshold:
                to_remove.append((n1, n2))
        self.graph_builder.G.remove_edges_from(to_remove)
        self.logger.info("Removed {} inefficient edges".format(len(to_remove)))

    def _score_syn_dna(
        self, n1: AssemblyNode, n2: AssemblyNode, edata: dict
    ) -> List[Edge]:
        if edata["type_def"].synthesize:
            span = edata["span"]
            if span > 0:
                # TODO: cyclic may not always be true
                region = self._edge_to_region(n1, n2)
                a = region.a
                c = region.c
                if c < a:
                    c += region.context_length
                assert c <= len(self.stats)
                complexity = self.stats.cost(a, c)
                if not complexity >= 0.0:
                    pass
                edata["complexity"] = complexity
                if self._complexity_to_efficiency(edata):
                    add_edge_note(edata, "highly_complex", True)
                    return True
        return False

    def score_complexity_edges(self, edges: List[Edge] = None) -> List[Edge]:
        """Score synthetic edges."""
        if edges is None:
            edges = self.graph.edges(data=True)
        complex_edges = []
        for n1, n2, edata in self.logger.tqdm(
            edges, "INFO", desc="Scoring synthetic DNA"
        ):
            if self._score_syn_dna(n1, n2, edata):
                complex_edges.append((n1, n2, edata))
        return complex_edges

    def update(self):
        self.logger.info("Post Processor: {}".format(self.stages))
        bad_edges = []

        self.logger.info("Scoring long PCR products")
        edges = list(self.graph.edges(data=True))

        if SequenceScoringConfig.SCORE_LONG in self.stages:
            for n1, n2, edata in self.logger.tqdm(
                edges, "INFO", desc="Scoring long PCR products"
            ):
                self.score_long_pcr_products(n1, n2, edata)
        if SequenceScoringConfig.SCORE_MISPRIMINGS in self.stages:
            for n1, n2, edata in self.logger.tqdm(
                edges, "INFO", desc="Scoring primer misprimings"
            ):
                self.score_primer_misprimings(n1, n2, edata)
        if SequenceScoringConfig.SCORE_COMPLEXITY in self.stages:
            bad_edges += self.score_complexity_edges(edges)
        self.logger.info(
            "Found {} highly complex synthesis segments".format(len(bad_edges))
        )

        if SequenceScoringConfig.PARTITION in self.stages:
            self.partition(bad_edges)

        # TODO: reimplement `remove_inefficient_edges`
        # self.remove_inefficient_edges()

    # TODO: add logging to graph post processor
    # TODO: partition gaps
    def __call__(self):
        self.logger.info("Post processing graph for {}".format(self.query.name))
        self.update()



from .designabc import DesignABC


class Design(DesignABC):
    def precompile(self):
        self.uncompile()
        with self.logger.timeit("DEBUG", "running blast"):
            self._blast()

    def postcompile(self, post_process_kwargs: dict = None):
        self.post_process_graphs(post_process_kwargs)



import functools
import hashlib
import json
import operator
from copy import deepcopy
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import networkx as nx

from dasi.constants import Constants as C
from dasi.exceptions import DasiOutputValidationError
from dasi.models import Molecule
from dasi.models import MoleculeType
from dasi.models import Reaction
from dasi.schemas import Schemas
from dasi.schemas import validate_with_schema
from dasi.utils.biopython import seqrecord_to_json


# TODO: Add "Order Primer" to list of reactions
def validate_output(data, do_raise: bool = True):
    return validate_with_schema(
        data,
        Schemas.output_schema,
        do_raise=do_raise,
        reraise_as=DasiOutputValidationError,
    )


def shash(seq: str) -> str:
    """Convert sequence string into a hash."""
    return hashlib.sha1(seq.strip().upper().encode()).hexdigest()


def mhash(mol1: Molecule) -> Tuple:
    """Convert molecule to hash."""
    return ("molecule", mol1.type.name, shash(str(mol1.sequence.seq)))


def rhash(rxn: Reaction, assembly_id: int) -> Tuple:
    """Convert reaction to hash."""
    outputs = sorted([mhash(m) for m in rxn.outputs])
    inputs = sorted([mhash(m) for m in rxn.inputs])
    if rxn.name == "Assembly":
        x = ["reaction", rxn.name, assembly_id]
    else:
        x = ["reaction", rxn.name]
    return tuple(x + [tuple(inputs)] + [tuple(outputs)])


def dasi_design_to_dag(
    design: Union["Design", "LibraryDesign"],
    validate: bool = True,
    elim_extra_reactions: bool = False,
    query_keys: Optional[List[str]] = None,
) -> nx.DiGraph:
    # TODO: standard way to display SeqRecord
    graph = nx.DiGraph()
    for q_i, (qk, result) in enumerate(design.results.items()):
        if query_keys and qk not in query_keys:
            continue
        for a_i, a in enumerate(result.assemblies):
            dk = (qk, a_i)
            for r in a.reactions:
                r1 = rhash(r, a_i)
                graph.add_node(r1, reaction=r, assembly_index=a_i, assembly=a)
                graph.nodes[r1].setdefault("design_keys", set()).add(dk)
                for in_mol in r.inputs:
                    n1 = mhash(in_mol)
                    graph.add_node(n1, molecule=in_mol)
                    graph.nodes[n1].setdefault("design_keys", set()).add(dk)
                    graph.add_edge(n1, r1)
                for out_mol in r.outputs:
                    n2 = mhash(out_mol)
                    graph.add_node(n2, molecule=out_mol)
                    graph.nodes[n2].setdefault("design_keys", set()).add(dk)
                    graph.add_edge(r1, n2)
    if validate:
        _validate_dag_graph(graph, elim_extra_reactions=elim_extra_reactions)
    return graph


def _validate_num_reactions(graph: nx.DiGraph, elim_extra_reactions: bool = False):
    """Validate that each DNA fragment for each plasmid has exactly one
    reaction. Optionally, remove these reactions from the final DAG.

    :param graph:
    :param elim_extra_reactions:
    :return:
    """
    nodes_to_remove = []
    for n2, ndata2 in graph.nodes(data=True):
        ########################
        # validate each non-plasmid is expected to
        # have at most one reaction
        ########################
        molecule = ndata2.get("molecule", None)

        if molecule and molecule.type is not MoleculeType.types[C.PLASMID]:
            predecessors = list(graph.predecessors(n2))
            if len(predecessors) > 1:
                if elim_extra_reactions:
                    for n1 in list(graph.predecessors(n2))[1:]:
                        nodes_to_remove.append(n1)
                else:
                    msg = str(n2)
                    for n1 in graph.predecessors(n2):
                        msg += "\n\t" + str(n1)
                    raise ValueError(
                        "Molecule {} has more than one reaction.\n{}".format(n2, msg)
                    )

    if nodes_to_remove:
        for n1 in nodes_to_remove:
            graph.remove_node(n1)


def _validate_reactions(graph):
    """Validate each reaction has expected number of predecessors and
    successors."""
    for n2, ndata2 in graph.nodes(data=True):
        reaction = ndata2.get("reaction", None)
        if reaction:
            predecessors = list(graph.predecessors(n2))
            if not len(predecessors) == len(reaction.inputs):
                raise ValueError(
                    "The number of predecessors ({}) does not"
                    " match the number of inputs for the reaction"
                    " ({})".format(len(predecessors), len(reaction.outputs))
                )

            successors = list(graph.successors(n2))
            if not len(successors) == len(reaction.outputs):
                raise ValueError(
                    "The number of successors ({}) does not"
                    " match the number of outputs for the reaction"
                    " ({})".format(len(successors), len(reaction.outputs))
                )


def _validate_dag_graph(graph: nx.DiGraph, elim_extra_reactions: bool = False):
    """Validates the DASi DAG graph."""
    _validate_num_reactions(graph, elim_extra_reactions)
    _validate_reactions(graph)


def _used_in_designs(ndata: Dict) -> Dict:
    x = sorted(list(ndata["design_keys"]))
    return [{"design_key": _x[0], "assembly": _x[1]} for _x in x]


def _reaction_metadata(reaction):
    return {
        "material_cost": reaction.metadata["material"],
        "efficiency": reaction.metadata.get("efficiency", 0.0),
        "complexity": reaction.metadata.get("complexity", None),
    }


def _assembly_metadata(assembly):
    rmetas = [_reaction_metadata(r) for r in assembly.nonassembly_reactions]
    material = sum([m["material_cost"] for m in rmetas])
    efficiency = functools.reduce(operator.mul, [m["efficiency"] for m in rmetas])
    complexities = [m["complexity"] for m in rmetas if m["complexity"] is not None]

    if complexities:
        max_complexity = max(complexities)
    else:
        max_complexity = 0

    return {
        "material_cost": material,
        "assembly_efficiency": efficiency,
        "max_complexity": max_complexity,
    }


def _reactions_property(
    graph: nx.DiGraph,
    reaction_node_dict: Dict[str, int],
    molecule_node_dict: Dict[str, int],
) -> Dict:
    property_reaction = []
    for n, index in reaction_node_dict.items():
        ndata = graph.nodes[n]
        reaction = ndata["reaction"]
        output_keys = sorted([mhash(m) for m in reaction.outputs])
        input_keys = sorted([mhash(m) for m in reaction.inputs])
        if reaction.name != Reaction.Types.Assembly:
            metadata = _reaction_metadata(reaction)
        else:
            assembly = ndata["assembly"]
            metadata = _assembly_metadata(assembly)

        property_reaction.append(
            {
                "__name__": ndata["reaction"].name,
                "__index__": index,
                "__type__": "reaction",
                "used_in_assemblies": _used_in_designs(ndata),
                "inputs": [molecule_node_dict[n] for n in input_keys],
                "outputs": [molecule_node_dict[n] for n in output_keys],
                "metadata": metadata,
            }
        )
    return property_reaction


def _clean_metadata(metadata):
    cleaned = {}
    for k, v in metadata.items():
        if isinstance(v, dict):
            cleaned[k] = _clean_metadata(v)
        else:
            try:
                json.dumps(v)
                cleaned[k] = v
            except Exception:
                pass
    return cleaned


def _molecules_property(
    graph: nx.DiGraph,
    reaction_node_dict: Dict[str, int],
    molecule_node_dict: Dict[str, int],
) -> Dict:
    property_molecule = []
    for n, index in molecule_node_dict.items():
        ndata = graph.nodes[n]
        mol = ndata["molecule"]
        used_as_inputs = [reaction_node_dict[n] for n in graph.successors(n)]
        used_as_outputs = [reaction_node_dict[n] for n in graph.predecessors(n)]
        property_molecule.append(
            {
                "__name__": mol.type.name,
                "__index__": index,
                "__type__": "molecule",
                "__meta__": deepcopy(_clean_metadata(mol.metadata)),
                "sequence": seqrecord_to_json(mol.sequence),
                "used_in_assemblies": _used_in_designs(ndata),
                "used_as_input_to_reactions": used_as_inputs,
                "used_as_output_to_reactions": used_as_outputs,
            }
        )
    return property_molecule


def _design_property(design, reaction_node_dict, graph):
    status = design.status

    def _reaction_summ(r, a_i):
        return {
            "reaction_index": reaction_node_dict.get(rhash(r, a_i)),
            "metadata": {
                "cost": r.metadata["cost"],
                "materials": r.metadata["cost"],
                "efficiency": r.metadata.get("efficiency", 0.0),
                "complexity": r.metadata.get("complexity", None),
            },
            "outputs": [
                {
                    "start": m.query_region.a,
                    "end": m.query_region.b,
                    "type": m.type.name,
                }
                for m in r.outputs
            ],
        }

    for qk, qstatus in status.items():
        qstatus["sequence"] = seqrecord_to_json(design.seqdb[qk])
        for a_i, adata in enumerate(qstatus["assemblies"]):
            assembly = design.results[qk].assemblies[a_i]

            nonassembly_reactions = [
                r for r in assembly.nonassembly_reactions if rhash(r, a_i) in graph
            ]
            assembly_reactions = [
                r for r in assembly.assembly_reactions if rhash(r, a_i) in graph
            ]

            adata["summary"] = [_reaction_summ(r, a_i) for r in nonassembly_reactions]
            adata["final_assembly_reaction"] = [
                reaction_node_dict[rhash(r, a_i)] for r in assembly_reactions
            ]

            # TODO: run start
    return status


def dasi_design_to_output_json(
    design: Union["Design", "LibraryDesign"],
    elim_extra_reactions: bool = False,
    query_keys: Optional[List[str]] = None,
):
    """Convert a DASi Design instance into an output JSON."""
    graph = dasi_design_to_dag(
        design, elim_extra_reactions=elim_extra_reactions, query_keys=query_keys
    )
    reaction_node_dict = {}
    molecule_node_dict = {}
    sorted_nodes = list(nx.topological_sort(graph))[::-1]
    for n in sorted_nodes:
        ndata = graph.nodes[n]
        r = ndata.get("reaction", None)
        m = ndata.get("molecule", None)
        if r:
            ndata["index"] = len(reaction_node_dict)
            assert n not in reaction_node_dict
            reaction_node_dict[n] = ndata["index"]
        elif m:
            ndata["index"] = len(molecule_node_dict)
            assert n not in molecule_node_dict
            molecule_node_dict[n] = ndata["index"]

            if m.type.name == C.PLASMID:
                pass
        else:
            raise ValueError

    output = {
        "metadata": design.metadata,
        "designs": _design_property(design, reaction_node_dict, graph),
        "molecules": _molecules_property(graph, reaction_node_dict, molecule_node_dict),
        "reactions": _reactions_property(graph, reaction_node_dict, molecule_node_dict),
    }

    OutputValidator.validate_design_assemblies(output)
    return output


def _filter_dict(data, keys):
    return {k: deepcopy(v) for k, v in data.items() if k in keys}


def filter_output_by_design_keys(
    output, keys, only_assemblies: Optional[List[str]] = None
):
    new_out = {}

    new_out["metadata"] = deepcopy(output["metadata"])
    new_out["designs"] = _filter_dict(output["designs"], keys)

    rids = []
    for design in new_out["designs"].values():
        for assembly in design["assemblies"]:
            for reaction in assembly["summary"]:
                rids.append(reaction["reaction_index"])
            rids += assembly["final_assembly_reaction"]
    rids = sorted(list(set(rids)))

    rdict = {r["__index__"]: r for r in output["reactions"]}
    mdict = {r["__index__"]: r for r in output["molecules"]}

    reactions = [rdict[r] for r in rids]

    mids = []
    for r in reactions:
        mids += r["inputs"]
        mids += r["outputs"]
    mids = sorted(list(set(mids)))

    molecules = [mdict[m] for m in mids]

    new_out["reactions"] = reactions
    new_out["molecules"] = molecules
    validate_output(new_out)
    return new_out


class OutputValidator:
    @classmethod
    def validate_design_assemblies(cls, out):
        for d in out["designs"].values():
            for a in d["assemblies"]:
                for rid in a["final_assembly_reaction"]:
                    reaction = out["reactions"][rid]
                    assert reaction["__name__"] == "Assembly"

    # TODO: In Silico from output json, compare to original sequences



r"""
NetworkX Utilities (:mod:`dasi.utils.networkx`)
=============================

.. currentmodule:: dasi.utils.networkx

This module provide various utility functions for networkx graphs

.. autosummary::
    :toctree: generated/

    algorithms
    exceptions
    shortest_path
    utils
"""
from .algorithms import floyd_warshall_with_efficiency
from .algorithms import sympy_floyd_warshall
from .exceptions import NetworkxUtilsException
from .shortest_path import sympy_dijkstras
from .shortest_path import sympy_multipoint_shortest_path
from .shortest_path import sympy_multisource_dijkstras



"""Alignments.

These classes represent abstract alignments between existing and
potential molecules that could be produced.
"""
import functools
import operator
from collections.abc import Sized
from typing import Dict
from typing import Generator
from typing import List
from typing import Union

from .molecule import MoleculeType
from dasi.exceptions import AlignmentException
from dasi.utils import argsorted
from dasi.utils import Region


class RepresentsMolecule:
    """Mixin for molecular sized alignments or alignment groups."""

    def __init__(self, query_region: Region, atype: str):
        self.query_region = query_region
        self.type = atype
        if atype not in MoleculeType.types:
            raise ValueError("atype '{}' not in MoleculeTypes".format(atype))

    def size_ok(self):
        """Determine if the size of this molecule is 'acceptable' from the
        bounded molecule type.

        :return: whether this passes the size requirement of the molecule type.
        """
        size = len(self.query_region)
        mol_type = MoleculeType.types[self.type]
        if mol_type.min_size is not None and size < mol_type.min_size:
            return False
        if mol_type.max_size is not None and size > mol_type.max_size:
            return False
        return True


class Alignment(RepresentsMolecule, Sized):
    """A pair of Regions that 'aligns' two regions of DNA sequences. All
    regions must always be the same length.

    A subregion of both regions may be taken.
    """

    __slots__ = [
        "query_region",
        "subject_region",
        "type",
        "query_key",
        "subject_key",
        "uid",
    ]

    def __init__(
        self,
        query_region: Region,
        subject_region: Region,
        atype: str,
        query_key: str,
        subject_key: str,
        meta: dict = None,
    ):
        """Makes an alignment between two regions of sequences. Validates the
        regions are the same length.

        :param query_region: Query region this alignment aligns to
        :param subject_region: Subject region this alignment aligns to.
        :param atype: Type of alignment
        :param query_key: The record identifier for the query
        :param subject_key: The record identifier for the subject
        """
        super().__init__(query_region, atype)
        assert query_region.direction == 1
        self.subject_region = subject_region
        self.validate()
        self.query_key = query_key
        self.subject_key = subject_key
        if meta is None:
            meta = {}
        self.meta = meta

    def validate(self):
        if not len(self.query_region) == len(self.subject_region):
            raise AlignmentException(
                "Regions must have the same size: {} vs {}. {} vs {}".format(
                    len(self.query_region),
                    len(self.subject_region),
                    self.query_region,
                    self.subject_region,
                )
            )

    def is_perfect_subject(self):
        return len(self.subject_region) == self.subject_region.context_length

    def sub_region(self, qstart: int, qend: int, atype=None) -> "Alignment":
        """Returns a copy of the alignment between the inclusive start and end
        relative to the query region.

        :param qstart: start of the query sub region
        :param qend: end of the query sub region
        :param atype: optional type of alignment to return
        :return:
        """
        query_copy = self.query_region.sub(qstart, qend)

        i = self.query_region.i(qstart)
        if i < 0:
            i = self.query_region.i(qstart + self.query_region.context_length)
        if i == len(self.subject_region):
            i = 0

        if self.subject_region.direction == -1:
            b = len(self.subject_region) - i
            a = b - len(query_copy)
            if a == b == len(self.subject_region):
                a = b = 0
            subject_copy = self.subject_region[a:b]
        else:
            subject_copy = self.subject_region[i : i + len(query_copy)]

        if atype is None:
            atype = self.type
        self.validate()
        return self.__class__(
            query_region=query_copy,
            subject_region=subject_copy,
            atype=atype,
            query_key=self.query_key,
            subject_key=self.subject_key,
        )

    def copy(self, atype=None) -> "Alignment":
        """Do shallow copy of this alignment. Query and subject regions between
        this and the copied alignment will be identical.

        :param atype: new alignment type
        :return:
        """
        if atype is None:
            atype = self.type
        return self.__class__(
            self.query_region,
            self.subject_region,
            atype,
            self.query_key,
            self.subject_key,
        )

    def __len__(self) -> int:
        return len(self.query_region)

    def __str__(self) -> str:
        return "<{} {} {} {}>".format(
            self.__class__.__name__, self.type, self.query_region, self.subject_region
        )

    @staticmethod
    def _rhash(region: Region):
        return (region.a, region.b, region.c, region.context_length, region.cyclic)

    def eq_hash(self):
        return (
            (self.query_key, self.subject_key, self.type)
            + self._rhash(self.query_region)
            + self._rhash(self.subject_region)
        )

    def __eq__(self, other: "Alignment"):
        return self.eq_hash() == other.eq_hash()

    def __repr__(self) -> str:
        return str(self)


class AlignmentGroupBase(RepresentsMolecule):
    """A representative Alignment representing a group of alignments."""

    __slots__ = ["query_region", "_alignments", "name", "type", "meta"]

    def __init__(
        self,
        alignments: List[Alignment],
        group_type: str,
        name: str = None,
        query_region: Region = None,
        meta: dict = None,
    ):
        """

        :param alignments:
        :param group_type:
        :param name:
        :param query_region:
        :param meta:
        """
        super().__init__(query_region, group_type)
        self._alignments = tuple(alignments)
        self.name = name
        if meta is None:
            meta = {}
        self.meta = meta

    @property
    def query_key(self) -> str:
        """Return the query key associated with the query region."""
        return self.alignments[0].query_key

    @property
    def subject_regions(self) -> List[Region]:
        """Return the list of subject regions in this alignment group."""
        return [a.subject_region for a in self.alignments]

    @property
    def subject_keys(self) -> List[str]:
        """Return the list of subject keys in this alignment group."""
        return [a.subject_key for a in self.alignments]

    @property
    def alignments(self):
        return self._alignments

    def sub_region(self, qstart: int, qend: int, atype: str) -> "AlignmentGroupBase":
        """Produce a new alignment group with sub-regions of the query region
        and subject regions at the specified new indicies."""
        alignments_copy = []
        for a in self.alignments:
            alignments_copy.append(a.sub_region(qstart, qend))
        for a in alignments_copy:
            a.type = atype
        return self.__class__(
            alignments=alignments_copy, group_type=atype, name="subregion"
        )

    def __repr__(self) -> str:
        return "<{} {}>".format(self.__class__.__name__, self.query_region)


class AlignmentGroup(AlignmentGroupBase):
    """A representative Alignment representing a group of alignments sharing
    the same starting and ending position for a query sequence."""

    __slots__ = list(AlignmentGroupBase.__slots__)

    def __init__(
        self,
        alignments: List[Alignment],
        group_type: str,
        name: str = None,
        meta: dict = None,
    ):
        super().__init__(
            alignments=alignments,
            group_type=group_type,
            name=name,
            query_region=alignments[0].query_region,
            meta=meta,
        )

    def reindex_alignments(self, indices: List[int]):
        """Reindex the alignments list by index.

        :param indices: list of indices
        :return: None
        """
        if len(indices) != len(self.alignments):
            raise ValueError(
                "Cannot reindex. Length of indices ({}) does not match length of"
                " alignments ({})".format(len(indices), len(self._groupings))
            )
        self._alignments = tuple(self._alignments[i] for i in indices)

    def prioritize_alignments(self, indices: List[int]):
        """Prioritize alignments by pushing groupings at the given indices into
        the front of the alignments list.

        :param indices: list of indices to prioritize
        :return: None
        """
        other_indices = []
        for i in range(len(self.alignments)):
            if i not in indices:
                other_indices.append(i)
        new_indices = indices + other_indices
        self.reindex_alignments(new_indices)


# class RepresentsPCR(AlignmentGroupBase):
#
#     @abstractmethod
#     def get_templates(self):
#         pass


class PCRProductAlignmentGroup(AlignmentGroupBase):
    """Represents a PCR product alignment from a template alignment and
    forward and reverse alignments. Represents several situations:

    ::

            Situations the PCRProductAlignmentGroup represents:

            Rev primer with overhang
                        <--------
            ------------------
            ---->

            Primers with overhangs
                           <--------
               ------------------
            -------->

            Primers 'within' the template
                    <-----
               ------------------
            -------->

            And so on...

    """

    __slots__ = list(set(AlignmentGroupBase.__slots__ + ["template", "fwd", "rev"]))

    def __init__(
        self,
        fwd: Union[None, Alignment],
        template: Alignment,
        rev: Union[None, Alignment],
        query_region: Region,
        group_type: str,
        meta: dict = None,
    ):
        """Initialize a new PCRProductAlignmentGroup. Represents a PCR product.
        Query region end points are determined from the first non-None
        alignment and the last non-None alignment. Produces *one new
        alignment*, the template alignment, which is the intersection of the
        provided template alignment and the query_region, which represents the
        exact region for which PCR primers ought to align in a PCR reaction.

        :param fwd: the forward primer alignment. Can be 'inside' the template or have
            an overhang.
        :param template: template alignment
        :param rev: the reverse primer alignment. Can be 'within' the template or have
            an overhang.
        :param group_type: group type name
        :param meta: extra meta data
        """
        if fwd is None and rev is None:
            raise AlignmentException("Must provide either a fwd and/or rev alignments")
        alignments = [x for x in [fwd, template, rev] if x is not None]

        a = alignments[0].query_region.a
        b = alignments[-1].query_region.b

        query_region = query_region.new(a, b)

        self.template = template
        self.fwd = fwd
        self.rev = rev

        super().__init__(
            alignments=alignments,
            group_type=group_type,
            query_region=query_region,
            meta=meta,
        )

    @property
    def subject_keys(self):
        raise AlignmentException(
            "Use subject keys directly, as in `self.fwd.subject_key`"
        )


# TODO: MultiPCRProductAlignmentGroup is a seriously convoluted class
#       Being such an important class, this should be very easy to understand.
#       `alignments` property should never be accessed directly, as
#          the concept of 'template' is different here (see `get_template`)
#
class MultiPCRProductAlignmentGroup(AlignmentGroupBase):
    """A PCR Product Alignment with redundant forward primer, reverse primer,
    and template alignments.

    Essentially, this represents region of a designed sequence *that could*
    be generated from a number of PCR reactions. Each PCR reaction is
    tracked in the `groupings`, which is a list of dictionaries with
    keys 'fwd', 'rev', 'template' and valued by Alignments.

    Now, there are alot of ways to produce PCR products.
    """

    __slots__ = list(set(AlignmentGroupBase.__slots__ + ["_groupings"]))

    EXPECTED_KEYS = "fwd", "rev", "template"
    TEMPLATE_ACCESSOR = "adjusted_template"

    def __init__(
        self,
        groupings: List[Dict[str, Alignment]],
        query_region: Region,
        group_type: str,
    ):
        """Initializes a new MultiPCRProductAlignmentGroup. This object
        represents a region of DNA that can be produced from a number of
        different forward, reverse, and template DNAs, all producing the same
        sequence.

        :param groupings: dictionary of groupings with the "fwd", "rev" and "template"
            keys.
        :param query_region: query region
        :param group_type: group type
        """

        for g in groupings:
            for key in self.EXPECTED_KEYS:
                if key not in g:
                    raise ValueError("Grouping is missing key '{}'".format(key))
        self._groupings = groupings
        self._templates = [None] * len(self._groupings)
        alignments = self._get_alignments()
        super().__init__(
            alignments=alignments, query_region=query_region, group_type=group_type
        )

    # @property
    # def groupings(self):
    #     return self._groupings

    def _get_alignments(self):
        accumulated = {}
        for key in self.EXPECTED_KEYS:
            accumulated.setdefault(key, list())
            for g in self._groupings:
                if g[key]:
                    accumulated[key].append(g[key])

        alignments = []
        for key in self.EXPECTED_KEYS:
            alignments += accumulated[key]

        return alignments

    def reindex_groupings(self, indices: List[int]):
        """Reindex the groupings list by index.

        :param indices: list of indices
        :return: None
        """
        if len(indices) != len(self._groupings):
            raise ValueError(
                "Cannot reindex. Length of indices ({}) does not match length of "
                "groups ({})".format(len(indices), len(self._groupings))
            )
        self._groupings = tuple(self._groupings[i] for i in indices)
        self._alignments = tuple(self._get_alignments())

    def prioritize_groupings(self, indices: List[int]):
        """Prioritize groupings by pushing groupings at the given indices into
        the front of the grouping list.

        :param indices: list of indices to prioritize
        :return: None
        """
        other_indices = []
        for i in range(len(self._groupings)):
            if i not in indices:
                other_indices.append(i)
        new_indices = indices + other_indices
        self.reindex_groupings(new_indices)

    # TODO: WHAT IS THIS METHOD???
    def get_template(self, index: int = 0):
        """Here we take the intersection of the template.query_region and query
        region. WHY???

        Notes this is **not** the 'template' key of the groupings.

        .. note::
            The result of this alignment is used in the primer design.

        :param index:
        :return:
        """
        group = self._groupings[index]

        if self.TEMPLATE_ACCESSOR not in group:
            template = self._groupings[index]["template"]
            intersection = template.query_region.intersection(self.query_region)
            group[self.TEMPLATE_ACCESSOR] = template.sub_region(
                intersection.a, intersection.b
            )
        return group[self.TEMPLATE_ACCESSOR]

    def iter_templates(self) -> Generator[Alignment, None, None]:
        """Generator of templates from `get_template`"""
        for i in range(len(self._groupings)):
            yield self.get_template(i)

    def get_fwd(self, index: int = 0) -> Alignment:
        """Get the forward alignment at the specified index.

        :param index: index of the grouping to access
        :return: alignment
        """
        group = self._groupings[index]
        return group["fwd"]

    def get_rev(self, index: int = 0):
        """Get the reverse alignment at the specified index.

        :param index: index of the grouping to access
        :return: alignment
        """
        group = self._groupings[index]
        return group["rev"]



"""sequence.

Methods for DNA/RNA sequence manipulation.
"""
from typing import List
from typing import Tuple

import numpy as np
from primer3plus.utils import anneal as p3panneal

BASES = "AGCT"


def random_sequence(size: int, letters=BASES) -> str:
    """Generate a random sequence string."""
    NP_BASES = np.array(list(letters))
    indices = np.random.randint(0, 4, size=size)
    bases = NP_BASES[indices]
    return "".join(bases)


def c(seq: str):
    """Return complement a dna string."""
    d = dict(zip("AGCTagct", "TCGAtcga"))
    return "".join([d[x] for x in seq])


def rc(seq: str):
    """Return a reverse complement a dna string."""
    return c(seq)[::-1]


def dna_like(seqstr: str, letters: str = "AGTCagctnNuU", min_length: int = 10):
    if seqstr is None:
        return False
    if len(seqstr) <= min_length:
        return False
    for s in seqstr:
        if s not in letters:
            return False
    return True


def anneal(
    template: str, primers: List[str], ignore_case: bool = True
) -> Tuple[List[dict], List[dict]]:
    assert isinstance(template, str)
    for p in primers:
        assert isinstance(p, str)
    return p3panneal(template, primers, ignore_case=ignore_case)



r"""
Utilities (:mod:`dasi.utils`)
=============================

This module provide various utility functions.

.. currentmodule:: dasi.utils

Utility modules
---------------

.. autosummary::
    :toctree: generated/

    npdf
    region
    sequence_design

Networkx utilities
------------------

.. currentmodule:: dasi.utils.networkx

Specialized networkx algorithms for path and cycle finding.

.. autosummary::
    :toctree: generated/

    algorithsm
    exceptions
    shortest_path
    utils

"""
import bisect
import inspect
from copy import deepcopy
from datetime import datetime
from functools import wraps
from itertools import tee
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union

from .networkx.shortest_path import multipoint_shortest_path
from .networkx.utils import sort_cycle
from .npdf import NumpyDataFrame
from .npdf import NumpyDataFrameException
from .region import Region


def sort_with_keys(a: Iterable[Any], key: Callable) -> Tuple[List, List]:
    """Sort an iterable, returning both the sorted array and the sorted keys.

    :param a: the iterable
    :param key: key function to use for sorting
    :return:
    """
    s = sorted(a, key=key)
    keys = [key(x) for x in s]
    return s, keys


def bisect_between(a: Iterable, low: Any, high: Any) -> Tuple[int, int]:
    """Returns the start (inclusive) and end (exclusive) indices for a sorted
    array using a key function.

    :param a: sorted array (does not check)
    :param low: low key
    :param high: high key
    :return: tuple of start (inclusive) and end (exclusive) indices
    """
    i = bisect.bisect_left(a, low)
    j = bisect.bisect_right(a[i:], high)
    return i, j + i


def bisect_slice_between(a: Iterable, keys: Iterable, low: Any, high: Any) -> Iterable:
    """Slice the iterable using inclusive bisection. Assumes both the iterable
    and keys are sorted. Bisect at specified `low` and `high`.

    :param a: pre-sorted iterable to slice
    :param keys: pre-sorted keys to bisect
    :param low: low key
    :param high: high key
    :return: sliced iterable
    """
    i, j = bisect_between(keys, low, high)
    return a[i:j]


def perfect_subject(data):
    """Determine whether a blast result consumes the entire subject."""
    if data["strand"] == 1 and data["start"] == 1 and data["raw_end"] == data["length"]:
        return True
    elif (
        data["strand"] == -1
        and data["raw_end"] == 1
        and data["start"] == data["length"]
    ):
        return True


def prep_df(df):
    colnames = [
        "DESIGN_ID",
        "DESIGN_KEY",
        "ASSEMBLY_ID",
        "REACTION_ID",
        "NAME",
        "TYPE",
        "KEY",
        "ROLE",
        "REGION",
        "SEQUENCE",
        "LENGTH",
        "META",
    ]
    df.columns = colnames
    df.sort_values(by=["TYPE", "DESIGN_ID", "REACTION_ID", "ASSEMBLY_ID", "ROLE"])
    return df


def group_by(arr: List[Any], key: Callable):
    """Group a list by some key."""
    grouped = {}
    for x in arr:
        k = key(x)
        grouped.setdefault(k, list())
        grouped[k].append(x)
    return grouped


def now():
    return datetime.now()


def log_times(key: str = None, class_attribute: str = "_method_trace"):
    """wrapper for logging method run times for a class."""

    def wrapped(f):
        @wraps(f)
        def _wrapped(self, *args, **kwargs):

            if not hasattr(self, class_attribute):
                raise ValueError(
                    "Instance {} must have attribute '{}'".format(self, class_attribute)
                )
            elif not isinstance(getattr(self, class_attribute), dict):
                raise ValueError(
                    "Attribute {} of {} must be a {}".format(
                        class_attribute, self, dict
                    )
                )
            t1 = now()
            result = f(self, *args, **kwargs)
            t2 = now()

            if key is None:
                use_key = f.__name__
            else:
                use_key = key
            getattr(self, class_attribute)[use_key] = (t1, t2)
            return result

        return _wrapped

    return wrapped


def fmt_datetime(t):
    return str(t)


def log_metadata(
    key: str = None,
    class_attribute: str = "_method_trace",
    additional_metadata: dict = None,
):
    """wrapper for logging method run times for a class."""

    def wrapped(f):
        @wraps(f)
        def _wrapped(self, *args, **kwargs):

            if not hasattr(self, class_attribute):
                raise ValueError(
                    "Instance {} must have attribute '{}'".format(self, class_attribute)
                )
            elif not isinstance(getattr(self, class_attribute), dict):
                raise ValueError(
                    "Attribute {} of {} must be a {}".format(
                        class_attribute, self, dict
                    )
                )
            t1 = now()
            result = f(self, *args, **kwargs)
            t2 = now()

            argspec = inspect.getfullargspec(f)

            copied_args = deepcopy(args)
            copied_kwargs = deepcopy(kwargs)
            argdict = dict(zip(argspec.args[1:], copied_args))
            argdict.update(copied_kwargs)

            metadata = {
                "__name__": f.__name__,
                "__spec__": str(argspec),
                "start": fmt_datetime(t1),
                "end": fmt_datetime(t2),
                "args": argdict,
            }
            if additional_metadata:
                metadata.update(additional_metadata)

            if key is None:
                use_key = f.__name__
            else:
                use_key = key
            getattr(self, class_attribute)[use_key] = metadata
            return result

        return _wrapped

    return wrapped


T = TypeVar("T")


def argsorted(
    arr: Iterable[T], key: Callable, return_items: bool = False
) -> Union[List[Tuple[int, T]], List[T]]:
    s = sorted(enumerate(tee(arr)[0]), key=lambda x: key(x[1]))
    if return_items:
        return s
    else:
        return [_s[0] for _s in s]


def lexsorted(keys: Iterable, target: Iterable[T], key: Callable) -> List[T]:
    sorted_indices = argsorted(enumerate(tee(keys)[0]), key=key)
    return [target[i] for i in sorted_indices]


def chunkify(arr: Iterable[T], chunk_size: int) -> Generator[List[T], None, None]:
    new_list = []
    for x in tee(arr, 1)[0]:
        new_list.append(x)
        if len(new_list) == chunk_size:
            yield new_list
            new_list = []
    if new_list:
        yield new_list



from itertools import product
from itertools import zip_longest

import networkx as nx
import numpy as np


def select_from_arrs(A, B, condition):
    """Returns an ndarray of same shape as A and B, selecting elements from
    either A or B according to the condition.

    .. note::     This method is called many thousands of times in the
    execution of     the modified floyd-warshall algorithm. This
    implementation is 2X faster     than whatever is implemented in
    `np.choose`, resulting in significant     speed improvements.
    """
    a = np.asarray(A).ravel()
    b = np.asarray(B).ravel()
    c = np.asarray(condition).ravel()
    d = c * len(a) + np.arange(len(b))
    e = np.hstack((a, b))
    return e[d].reshape(A.shape)


def replace_nan_with_inf(m):
    np.putmask(m, np.isnan(m), np.full_like(m, np.inf))
    return m


def divide(mata, matb):
    """Divides two matricies, replacing NaN with np.inf."""
    matc = mata / matb
    replace_nan_with_inf(matc)
    return matc


def compare_matrix_to_true_paths(mat, path_dict, nodelist):
    node_to_index = {n: i for i, n in enumerate(nodelist)}
    for n1 in path_dict:
        for n2, edata in path_dict[n1].items():
            u = node_to_index[n1]
            v = node_to_index[n2]

            min_cost = edata["cost"]
            est_cost = mat[u, v]
            if min_cost != est_cost:
                status = "ERROR {}".format(min_cost - est_cost)
            else:
                status = ""
            print(
                "{} -> {:5}: {:10.2f} {:13.2f} {:25}".format(
                    u, v, min_cost, est_cost, status
                )
            )


def min_edit_distance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1
    distances = range(len(s1) + 1)
    for index2, char2 in enumerate(s2):
        newDistances = [index2 + 1]
        for index1, char1 in enumerate(s1):
            if char1 == char2:
                newDistances.append(distances[index1])
            else:
                newDistances.append(
                    1
                    + min((distances[index1], distances[index1 + 1], newDistances[-1]))
                )
        distances = newDistances
    return distances[-1]


def compare_path_rank_order(mat, path_dict, nodelist, verbose=False):
    # get sorted index pairs representing end to end shortest paths
    indices = np.asarray(mat).flatten().argsort().flatten()
    index_pairs = np.dstack(np.unravel_index(indices, mat.shape))
    y = index_pairs.reshape(index_pairs.shape[1], index_pairs.shape[2])
    y = [(mat[u, v], u, v) for u, v in y]
    matrix_shortest_index_pairs = [(u, v) for c, u, v in sorted(y)]

    x = []
    node_to_index = {n: i for i, n in enumerate(nodelist)}
    for n1 in path_dict:
        for n2, edata in path_dict[n1].items():
            u = node_to_index[n1]
            v = node_to_index[n2]
            x.append((edata["cost"], u, v))
    true_shortest_index_pairs = [(u, v) for c, u, v in sorted(x)]

    edit_dist = min_edit_distance(
        true_shortest_index_pairs, matrix_shortest_index_pairs
    )
    print("Distance: {}".format(edit_dist))
    if verbose:
        for true_pair, mat_pair in zip_longest(
            true_shortest_index_pairs, matrix_shortest_index_pairs
        ):
            print("{}  {}".format(true_pair, mat_pair))


def sort_cycle(arr, key=None):
    """Sort a cyclic array, maintaining order."""
    if key is None:
        arr_with_i = sorted([(x, i) for i, x in enumerate(arr)])
    else:
        arr_with_i = sorted([(key(x), i) for i, x in enumerate(arr)])
    i = arr_with_i[0][1]
    return arr[i:] + arr[:i]



import hashlib
import os
from glob import glob

from dasi import cost


def hashfiles(files, hash="sha1", hashfunc=None):

    if not hashfunc:

        def hashfunc(string):
            return getattr(hashlib, hash)(string.encode("utf-8")).hexdigest()

    contents = ""
    sorted_path = sorted(files)
    for file in sorted_path:
        with open(file, "r") as f:
            contents += hashfunc(f.read())
    return hashfunc(contents)


def cost_module_checksum():
    cost_dir = os.path.dirname(cost.__file__)
    cost_files = sorted(
        glob(os.path.join(cost_dir, "*.py")) + glob(os.path.join(cost_dir, "*.json"))
    )
    return hashfiles(cost_files)


def cached(path, save_func, load_func, checksum_path, logger=None):
    different_checksum = True
    checksum = cost_module_checksum()
    if os.path.isfile(checksum_path):
        with open(checksum_path, "r") as f:
            stored_checksum = f.read().strip()
            if stored_checksum == checksum:
                different_checksum = False
            if logger:
                logger.debug("Stored checksum: {}".format(stored_checksum))
    if logger:
        logger.debug("Checksum: {}".format(checksum))
    if different_checksum or not os.path.isfile(path):
        if logger:
            logger.debug("Using default params")
        model = save_func(path)
        with open(checksum_path, "w") as f:
            f.write(checksum)
        stat = os.stat(path)
        if logger:
            logger.debug("Wrote {} bytes".format(stat.st_size))
    else:
        if logger:
            logger.debug("Loading {}".format(path))
        stat = os.stat(path)
        if logger:
            logger.debug("Loaded {} bytes".format(stat.st_size))
        model = load_func(path)
    return model



import functools
from functools import lru_cache
from itertools import repeat
from itertools import zip_longest
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

import numpy as np

from dasi.utils.sequence.sequence_complexity import DNAStats


@lru_cache(254)
def cached_stats_cost(stats: DNAStats, i: int, j: int):
    return stats.cost(i, j)


def stats_cost_arr(stats: DNAStats, i: Iterable[int], j: Iterable[int]):
    for _i, _j in zip_longest(i, j, fillvalue="STOP"):
        if i == "STOP":
            raise ValueError("Iterables are different sizes.")
        elif j == "STOP":
            raise ValueError("Iterables are different sizes")
        yield cached_stats_cost(stats, _i, _j)


def get_partitions(stats: DNAStats, x: List[int], delta: int = 0):
    """Get partition costs for a list of partition indices."""
    x1 = np.array(x)
    if delta is None:
        x2 = x1
    else:
        x2 = x1 + delta
    c1 = np.array(list(stats_cost_arr(stats, repeat(None, len(x1)), x1)))
    c2 = np.array(list(stats_cost_arr(stats, x2, repeat(None, (len(x2))))))
    return x1, c1, c2


def _nearest_step(x, step_size):
    return ((x - 1) // step_size + 1) * step_size


def find_opt_partition(
    stats,
    i: Optional[Union[None, int]] = None,
    j: Optional[Union[None, int]] = None,
    step_size: int = 10,
    delta: Optional[int] = None,
    use_nearest_step: bool = False,
):
    """Find the optimal partition in range (i, j) given the stepsize."""

    # readjust positions to nearest step size to benefit from cache
    if i is None:
        i = 0
    elif use_nearest_step:
        i = _nearest_step(i, step_size)
    if j is None:
        j = len(stats.seq)
    elif use_nearest_step:
        j = min(len(stats.seq), _nearest_step(j, step_size))

    x = np.arange(i, j, step_size, dtype=np.int32)
    _, c1, c2 = get_partitions(stats, x, delta=delta)
    c = c1 + c2
    non_nan_filter = ~np.isnan(c)
    x = x[non_nan_filter]
    c = c[non_nan_filter]

    if not len(c):
        if delta is None:
            return None, None
        else:
            return None, None, None
    cmin = c.min()
    i = np.argmin(c)
    return x[i], cmin


def find_fast_opt_partition(
    stats,
    i=None,
    j=None,
    step_size: int = 100,
    threshold: int = 10,
    delta: Optional[int] = None,
    step_size_refinement: int = 10,
):
    f = functools.partial(find_opt_partition, delta=delta)
    p, pmin = f(stats, i=i, j=j, step_size=step_size)
    if p is None:
        return None, None
    if pmin < threshold:
        return p, pmin
    p2, p2min = f(
        stats,
        max(0, p - step_size),
        min(p + step_size, len(stats.seq)),
        step_size=int(step_size / step_size_refinement),
    )
    if p2 is None:
        return p, pmin
    else:
        return p2, p2min


def find_best_partitions(
    stats: DNAStats,
    threshold: int,
    i=None,
    j=None,
    step_size: int = 100,
    delta: Optional[int] = None,
    partitions=None,
):
    if partitions is None:
        partitions = []
    c = stats.cost(i, j)
    if c < threshold:
        return partitions
    p, pmin = find_fast_opt_partition(stats, i=i, j=j, step_size=step_size, delta=delta)
    if p is None or p in partitions:
        return partitions
    else:
        partitions.append(p)
    c1 = stats.cost(i, p)
    c2 = stats.cost(p, j)
    if c1 > threshold:
        find_best_partitions(
            stats, threshold=threshold, i=i, j=p, partitions=partitions
        )
    if c2 > threshold:
        find_best_partitions(
            stats, threshold=threshold, i=p, j=j, partitions=partitions
        )

    return partitions


def _shift_indices(indices: List[int], origin: int, length: int):
    shifted_indices = []
    for i in indices:
        i += i + origin
        if i >= length:
            i -= length
        shifted_indices.append(i)
    return shifted_indices


def find_by_partitions_for_sequence(
    stats: DNAStats,
    cyclic: bool,
    threshold: int,
    step_size: int = 100,
    delta: Optional[int] = None,
):
    """Approximates the best partitions for a sequence. If cyclic=True, then
    will approximate partitions by also rotating the origin.

    :param stats: DNAStats instance
    :param cyclic: whether the sequence is cyclic
    :param threshold: threshold cost to find a partition
    :param step_size: step size to find partition.
    :return:
    """
    f = functools.partial(
        find_best_partitions, threshold=threshold, step_size=step_size, delta=delta
    )
    partitions = f(stats)
    if cyclic:
        origin = int(len(stats.seq) / 2.0)
        seq = stats.seq
        stats2 = stats.copy_with_new_seq(seq[origin:] + seq[:origin])
        partitions2 = f(stats2)
        partitions += _shift_indices(partitions2, origin, len(seq))
    partitions = sorted(set(partitions))
    return partitions



"""Module for calculating the complexity of a DNASequence.

Complexity rules borrowed from IDT.

* less than 25% GC content
* higher than 75% GC content
* homopolymeric runs of 10 or more A/Ts
* homopolymeric runs of 6 or more G/Cs
* repeats greater than 14bp
* hairpins
* GC scew?
* windowed scew?
"""
import hashlib
import json
from copy import copy as do_copy
from copy import deepcopy as do_deepcopy
from functools import lru_cache

import numpy as np


class DNAStatsMeta:
    def __init__(self):
        self.fwd_signatures = None
        self.rev_signatures = None
        self.rolling_stats = None
        self.repeat_signatures = None


# TODO:' compute cost for extremes of GC content
# TODO: this could be a separate library
class DNAStats:
    """
    Class for computing statistics for DNA Sequences.

    ::

        {
            "n_repeats",
            "n_hairpins",
            "window_cost",
            "gc_cost"
        }
    """

    BASES = "AGCT"
    np.random.seed(1)
    BASE_SIGNATURES = np.random.uniform(0.0, 1.0, size=len(BASES)).reshape(-1, 4)

    DEFAULT_MODE = ("window", "hairpin", "repeat")
    ONLY_WINDOW = ("window",)
    ONLY_HAIRPIN = ("hairpin",)
    ONLY_REPEAT = "repeat"

    __slicable__ = [
        "fwd_signatures",
        "rev_signatures",
        "repeat_signatures",
        "seq",
        "seq_onehot",
        "rolling_stats",
    ]

    __copyable__ = [
        "bases",
        "repeat_window",
        "stats_window",
        "hairpin_window",
        "gc_content_threshold",
        "at_content_threshold",
        "base_percentage_threshold",
        "conv_seed_repeat",
        "mode",
        "conv_seed_hairpin",
    ]

    __slots__ = __copyable__ + __slicable__

    def __init__(
        self,
        seq,
        repeat_window,
        stats_window,
        hairpin_window,
        base_percentage_threshold: float = 0.75,
        gc_content_threshold: float = 0.85,
        at_content_threshold: float = 0.90,
        conv_seed_repeat: np.array = None,
        conv_seed_hairpin: np.array = None,
        mode: str = None,
    ):
        """

        :param seq:
        :param repeat_window: the length of the k-mers to search for repeats.
        :param stats_window: the width of the stats window
        :param hairpin_window: length of hte k-mers to search for hairpins.
        :param base_percentage_threshold: if the per base percentage threshold
            to count in slinding window calculatiions. For example,
            for a sliding window of 20, if the threshold is 0.75, if 15 bases
            or of a single base pair, this will get counted in the `window_cost`
            method.
        :param gc_content_threshold: if the gc percentage threshold
            to count in sliding window calculatiions. For example,
            for a sliding window of 20, if the threshold is 0.75, if 15 bases
            of any window of 20 were gc, this will get counted in the `window_cost`
            method.
        :param conv_seed_repeat: (optional) the array to convolve the one-hot
            encoded DNA sequence with. If not provided, will be generated by
            `np.random.uniform(0.0, 100.0, size=repeat_window)`
        :param conv_seed_hairpin: (optional) the array to convolve the one-hot
            encoded DNA sequence with to find hairpin signatures.
            If not provided, will be generated by
            `np.random.uniform(0.0, 100.0, size=hairpin_window)`
        :param mode: (optional) Set to DNAStats.ONLY_HAIRPIN to only
            calculate hairpin signatures, DNAStats.ONLY_REPEATS to only calculate
            repeat signatures. DNAStats.ONLY_WINDOW to perform only
            sliding window calculations.
        """
        if not isinstance(seq, str):
            raise ValueError("Expected a str not a '{}'".format(type(seq)))
        seq = seq.upper()
        self.seq = seq
        arr = np.array(list(self.seq))
        self.bases = self.BASES.upper()
        self.seq_onehot = self.one_hot(arr, self.bases)
        self.repeat_window = repeat_window
        self.stats_window = stats_window
        self.hairpin_window = hairpin_window
        self.gc_content_threshold = gc_content_threshold
        self.at_content_threshold = at_content_threshold
        self.base_percentage_threshold = base_percentage_threshold
        if not mode:
            mode = self.DEFAULT_MODE
        self.mode = mode
        if conv_seed_repeat is not None:
            self.conv_seed_repeat = conv_seed_repeat
        else:
            self.conv_seed_repeat = np.random.uniform(0.0, 100.0, size=repeat_window)
        if conv_seed_hairpin is not None:
            self.conv_seed_hairpin = conv_seed_hairpin
        else:
            self.conv_seed_hairpin = np.random.uniform(0.0, 100.0, size=hairpin_window)

        if "window" in self.mode:
            rolling_stats = self.get_base_stats(stats_window)
            gc_content = (
                rolling_stats[self.BASES.index("G"), :]
                + rolling_stats[self.BASES.index("C"), :]
            )
            at_content = (
                rolling_stats[self.BASES.index("A"), :]
                + rolling_stats[self.BASES.index("T"), :]
            )
            self.rolling_stats = np.vstack((rolling_stats, gc_content, at_content))
        else:
            self.rolling_stats = None

        ###
        # get repeat signatures
        ###
        seq_signatures = np.sum(
            np.multiply(self.seq_onehot, self.BASE_SIGNATURES.T), axis=0
        )
        if "repeat" in self.mode:
            mv = np.convolve(seq_signatures, self.conv_seed_repeat, mode="valid")
            self.repeat_signatures = np.concatenate(
                ([np.NaN for _ in range(repeat_window - 1)], mv)
            )
        else:
            self.repeat_signatures = None

        ###
        # get hairpin signatures
        ###
        if "hairpin" in self.mode:
            revcomp_signatures = np.sum(
                np.multiply(self.seq_onehot[:, ::-1], self.BASE_SIGNATURES[:, ::-1].T),
                axis=0,
            )
            self.fwd_signatures = np.convolve(
                seq_signatures, self.conv_seed_hairpin, mode="valid"
            )
            self.rev_signatures = np.convolve(
                revcomp_signatures, self.conv_seed_hairpin, mode="valid"
            )[::-1]
        else:
            self.fwd_signatures = None
            self.rev_signatures = None

    def _view(self, index: slice = None, copy: bool = False, deepcopy: bool = False):
        x = object.__new__(self.__class__)
        for k in self.__slots__:
            v = getattr(self, k)
            if k in self.__slicable__:
                if copy or deepcopy:
                    if isinstance(v, np.ndarray):
                        v = v.copy()
                    else:
                        v = do_copy(v)
                if index:
                    if isinstance(v, np.ndarray):
                        if v.ndim == 1:
                            v = v[index]
                        else:
                            v = v[:, index]
                    else:
                        v = v[index]
            elif k in self.__copyable__:
                if deepcopy:
                    v = do_deepcopy(getattr(self, k))
                else:
                    v = do_copy(getattr(self, k))
            setattr(x, k, v)
        return x

    def copy(self, index: slice = None):
        return self._view(index=index, copy=True)

    def deepcopy(self, index: slice = None):
        return self._view(index=index, deepcopy=True)

    def view(self, index: slice = None):
        return self._view(index=index)

    @property
    def config(self):
        return dict(
            repeat_window=self.repeat_window,
            stats_window=self.stats_window,
            hairpin_window=self.hairpin_window,
            base_percentage_threshold=self.base_percentage_threshold,
            gc_content_threshold=self.gc_content_threshold,
            at_content_threshold=self.at_content_threshold,
            mode=self.mode,
        )

    def copy_with_new_seq(self, seq: str):
        return self.__class__(
            seq,
            repeat_window=self.repeat_window,
            stats_window=self.stats_window,
            hairpin_window=self.hairpin_window,
            base_percentage_threshold=self.base_percentage_threshold,
            gc_content_threshold=self.gc_content_threshold,
            at_content_threshold=self.at_content_threshold,
            conv_seed_repeat=self.conv_seed_repeat,
            conv_seed_hairpin=self.conv_seed_hairpin,
            mode=self.mode,
        )

    @staticmethod
    def one_hot(sequence, categories):
        arrs = []
        for i, c in enumerate(categories):
            a = sequence == c
            a = a.astype(int)
            arrs.append(a)
        return np.vstack(arrs)

    @staticmethod
    def rolling_window(a, window):
        shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
        strides = a.strides + (a.strides[-1],)
        return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

    @staticmethod
    def rolling_average(x, n):
        mv = np.convolve(x, np.ones(n) / n, mode="valid")
        return np.concatenate(([np.NaN for k in range(n - 1)], mv))

    def get_base_stats(self, window):
        return np.mean(self.rolling_window(self.seq_onehot, window), axis=2)

    def get_repeat_signatures(self, window):
        seq_signatures = np.sum(
            np.multiply(self.seq_onehot, self.BASE_SIGNATURES.T), axis=0
        )
        x = np.random.uniform(0.0, 100.0, size=window)
        mv = np.convolve(seq_signatures, x, mode="valid")
        signatures = np.concatenate(([np.NaN for _ in range(window - 1)], mv))
        return signatures

    def get_hairpin_signatures(self, window):
        seq_signatures = np.sum(
            np.multiply(self.seq_onehot, self.BASE_SIGNATURES.T), axis=0
        )
        revcomp_signatures = np.sum(
            np.multiply(self.seq_onehot[:, ::-1], self.BASE_SIGNATURES[:, ::-1].T),
            axis=0,
        )
        x = np.random.uniform(0.0, 100.0, size=window)
        mv1 = np.convolve(seq_signatures, x, mode="valid")
        mv2 = np.convolve(revcomp_signatures, x, mode="valid")
        return mv1, mv2[::-1]

    @staticmethod
    def count_signatures(a):
        return len(np.where(np.unique(a, return_counts=True)[1] > 1)[0])

    def slice_repeats(self, i, j):
        return self.count_signatures(self.repeat_signatures[i:j])

    def slice_hairpins(self, i, j):
        f = self.fwd_signatures[i:j]
        r = self.rev_signatures[i:j]
        d = np.concatenate((f, r))
        num_hairpins = self.count_signatures(d) - 2 * self.count_signatures(f)
        return num_hairpins

    def window_cost(
        self,
        i,
        j,
        base_threshold_perc: float = None,
        gc_content_threshold: float = None,
        at_content_threshold: float = None,
    ):

        if base_threshold_perc is None:
            base_threshold_perc = self.base_percentage_threshold
        if gc_content_threshold is None:
            gc_content_threshold = self.gc_content_threshold

        if at_content_threshold is None:
            at_content_threshold = self.at_content_threshold
        d = self.rolling_stats[:4, i:j]

        # keep each base below 0.8
        a = np.sum(d > base_threshold_perc, axis=1)

        # keep gc content below extremes 0.7
        b = np.sum(self.rolling_stats[4, i:j] > gc_content_threshold)

        # keep gc content below extremes 0.7
        c = np.sum(self.rolling_stats[5, i:j] > at_content_threshold)
        return a, b, c

    # def get_GC_content_complexity(seq):
    #     gc = get_GC_content(seq)
    #     return abs(gc * 100.0 - 50) * 17 / 25.0

    @staticmethod
    def _optimize_partition_helper(
        signatures: np.ndarray, step: int, i: int = None, j: int = None
    ):
        """Optimize partition by minimizing the number of signatures in the
        given array.

        :param signatures: array of signatures
        :param step: step size
        :param i:
        :param j:
        :return:
        """
        d = []

        if i is None:
            i = 0
        if j is None:
            j = signatures.shape[1]

        for x in range(i, j, step):
            m1 = np.empty(signatures.shape[1])
            m2 = m1.copy()
            m1.fill(np.nan)
            m2.fill(np.nan)

            m1[:x] = np.random.uniform(1, 10)
            m2[x:] = np.random.uniform(1, 10)

            d += [m1, m2]
        d = np.vstack(d)
        z = np.tile(d, signatures.shape[0]) * signatures.flatten()

        partition_index = np.repeat(
            np.arange(0, signatures.shape[1], step),
            signatures.shape[0] * signatures.shape[1] * 2,
        )

        a, b, c = np.unique(z, return_counts=True, return_index=True)
        i = b[np.where(c > 1)]
        a, c = np.unique(partition_index[i], return_counts=True)
        if len(c):
            arg = c.argmin()
            return a[arg], c[arg]

    def count_repeats_from_slice(self, i, j):
        """Counts the number of repeats found in slices [0, i) and [j, None) of
        all kmers in the sequence found within the [i, j) slice.

        .. note::
            The hairpin_window_length of the DNAStats instance is used as the kmer
            length.

        :param i: start of the slice (inclusive)
        :param j: end of the slice (exclusive)
        :return: tuple of repeat counts found in left and right slices respectively.
        """
        if i < 0:
            raise IndexError("i={} cannot be < 0")
        if j > len(self.seq):
            raise IndexError("j={} cannot be >= length {}".format(j, len(self.seq)))
        kmer_length = self.hairpin_window

        # we ust signatures (src) of kmers inside i:j...
        src_indices = np.arange(i, j - kmer_length + 1)

        # too look for signatures (dest) in the left and right flanks
        left_flank_indices = np.arange(0, i)
        right_flank_indices = np.arange(j, len(self.seq) - kmer_length)
        dest_indices = np.concatenate([left_flank_indices, right_flank_indices])

        src = self.fwd_signatures[src_indices]

        # TODO: unsure why the rev indices need to be shifted...
        src_rc = self.rev_signatures[src_indices]
        dest = self.fwd_signatures[dest_indices]

        def _mispriming_counts(src_sig, src_sig_rc, dest_sig):
            # 5' end, top repeats to self
            counts_trim_top = self.count_signatures(src_sig)

            # 5' end, calculate priming to top strand
            x, counts = np.unique(src_sig, return_counts=True)
            top_counts = counts[np.where(np.isin(x, dest_sig))].sum()

            # 5' end, calculate priming to bottom strand
            x, counts = np.unique(src_sig_rc, return_counts=True)
            bottom_counts = (
                counts[np.where(np.isin(x, dest_sig))].sum() + (counts - 1).sum()
            )

            return top_counts + bottom_counts + counts_trim_top

        return _mispriming_counts(src, src_rc, dest)

    def cost(self, i=None, j=None):
        d = self(i, j)
        w1 = d["window_cost"][0]
        w2 = d["window_cost"][1]
        total = (
            d["n_repeats"] + d["n_hairpins"] + np.sum(w1) + np.sum(w2) + d["gc_cost"]
        )
        return total

    @staticmethod
    def _hash(string: str):
        return hashlib.sha1(string.encode("utf-8")).hexdigest()

    def __hash__(self):
        config = json.dumps(self.config)
        s = self.seq.upper() + config
        return int(hashlib.sha1(s.encode("utf-8")).hexdigest(), 16)

    def gc_cost(self, i, j):
        mn = np.mean(self.seq_onehot[:, i:j], axis=1)
        gi = self.bases.index("G")
        ci = self.bases.index("C")
        gc = mn[gi] + mn[ci]
        gccost = abs(gc * 100.0 - 50) * 17 / 25.0
        return gccost

    def __call__(self, i=None, j=None):
        return {
            "n_repeats": self.slice_repeats(i, j),
            "n_hairpins": self.slice_hairpins(i, j),
            "window_cost": self.window_cost(i, j),
            "gc_cost": self.gc_cost(i, j),
        }

    def __len__(self):
        return len(self.seq)

    def __getitem__(self, item):
        return self.view(item)


@lru_cache(512)
def cached_stats(seq: str, hairpin_window: int):
    return DNAStats(seq, 1, 1, hairpin_window, mode=DNAStats.ONLY_HAIRPIN)


@lru_cache(512)
def count_misprimings_in_amplicon(
    seq,
    i,
    j,
    min_primer_anneal: int = 12,
    max_primer_anneal: int = 30,
    cyclic: bool = False,
    max_length: int = 3000,
):
    """Counts the estimated number of misprimings in an amplicon.

    :param seq: The template sequence
    :param i: amplicon start (inclusive)
    :param j: amplicon end (exclusive)
    :param min_primer_anneal: The minimum number of bases to look for misprimings
        (exact matches only).
    :param max_primer_anneal: The maximum size of the annealing sequence of the primers.
        Internally, this takes a window slice from the ends of the amplicon to look
        for repeats in the template.
    :param cyclic: If True, assumes the sequence is cyclic.
    :param max_length: The maximum length to consider a mispriming. Generally,
        this should be about the bp distinguishable in your fragment purification
        (e.g. gel)
    :return:
    """
    if j == i:
        return 0
    if i < 0:
        raise IndexError("i cannot be < 0")
    if j < 0:
        raise IndexError("j cannot be < 0")
    elif j < i and not cyclic:
        raise IndexError(
            "Invalid indices provided. "
            "Indices [{i},j) indicates a cyclic sequence, but "
            "cyclic was set to False."
        )
    elif cyclic:
        seq = seq[i:] + seq[:i]
        if j < i:
            length = len(seq) - i + j
            i = 0
            j = length
        else:
            j = j - i
            i = 0

        # TODO: trim extra long sequences
        # trim extra long sequences
        # if len(seq) - j > 2 * max_length:
        #     seq = seq[:j] + seq[j:][max_length] + seq[-max_length:]
    # else:
    #     # trim extra long sequences
    #     if i > max_length:
    #         delta = max_length - i
    #     else:
    #         delta = 0
    #     seq = seq[delta:i] + seq[i:j] + seq[j:][:max_length]
    #     if i > max_length:
    #         delta = max_length - i
    #         i -= delta
    #         j -= delta
    stats = cached_stats(seq, min_primer_anneal)
    n1 = stats.count_repeats_from_slice(i, min(i + max_primer_anneal, len(seq) - 1))
    k = j - max_primer_anneal
    if k < 0:
        n2 = 0
    else:
        n2 = stats.count_repeats_from_slice(max(j - max_primer_anneal, 0), j)
    return n1 + n2



r"""
Design (:mod:`dasi.design`)
=============================

.. currentmodule:: dasi.design

This module provide DNA assembly functionality for DASi.
"""
import functools
import operator
from abc import ABC
from abc import abstractmethod
from multiprocessing import Pool
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import networkx as nx
import pandas as pd
from Bio.SeqRecord import SeqRecord
from more_itertools import pairwise
from pyblast import BioBlastFactory
from pyblast.utils import is_circular

from .design_result import DesignResult
from .graph_builder import AssemblyGraphBuilder
from .graph_builder import AssemblyGraphPostProcessor
from .optimize import optimize_graph
from dasi.__version__ import __title__
from dasi.__version__ import __version__
from dasi.constants import Constants
from dasi.cost import cached_span_cost
from dasi.cost import SpanCost
from dasi.design.graph_builder import AssemblyNode
from dasi.design.output import dasi_design_to_output_json
from dasi.design.output import validate_output
from dasi.design.report import Report
from dasi.exceptions import DasiDesignException
from dasi.log import logger
from dasi.models import Alignment
from dasi.models import AlignmentContainer
from dasi.models import AlignmentContainerFactory
from dasi.schemas import Schemas
from dasi.utils import chunkify
from dasi.utils import log_metadata
from dasi.utils import perfect_subject
from dasi.utils.sequence import generate_fake_designs

BLAST_PENALTY_CONFIG = {
    "gapopen": 3,
    "gapextend": 3,
    "reward": 1,
    "penalty": -5,
    "ungapped": None,
}


def assemble_graph(
    container: AlignmentContainer, span_cost: SpanCost
) -> Tuple[nx.DiGraph, AlignmentContainer]:
    """Build an assembly graph for a specified query."""
    container.expand(expand_overlaps=True, expand_primers=True)
    container.clean_alignments()
    container.groups()
    container.freeze()
    graph_builder = AssemblyGraphBuilder(container, span_cost=span_cost)
    graph = graph_builder.build_assembly_graph()
    return graph, container


class FakePool:
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return FakePool()

    def __exit__(self, a, b, c):
        pass

    @staticmethod
    def map(func, args):
        return [func(arg) for arg in args]


class DesignABC(ABC):
    """Design class that returns optimal assemblies from a set of materials."""

    PRIMERS = "primers"
    TEMPLATES = "templates"
    QUERIES = "queries"
    FRAGMENTS = "fragments"
    DEFAULT_N_ASSEMBLIES = 3
    DEFAULT_JOB_SIZE = 1
    DEFAULT_N_JOBS = 1
    ALGORITHM = Constants.ALGORITHM_DEFAULT

    def __init__(self, span_cost=None, seqdb=None, name: str = ""):
        """

        :param span_cost:
        :type span_cost: SpanCost
        :param seqdb:
        :type seqdb: dict
        """
        self.name = name
        self.blast_factory = BioBlastFactory(config=BLAST_PENALTY_CONFIG)
        self.logger = logger(self)

        # graph by query_key
        if seqdb is None:
            seqdb = {}
        self._seqdb = seqdb  #: Sequence dict registry
        if span_cost is None:
            span_cost = cached_span_cost()
        self.span_cost = span_cost  #: span cost df
        self.graphs = {}  #: Dict[str, nx.DiGraph]
        self._results = {}  #: Dict[str, DesignResult]
        self.template_results = []
        self.fragment_results = []
        self.primer_results = []
        self.container_factory = AlignmentContainerFactory(self.seqdb)
        self._times = {}
        self._method_trace = {}

    def _get_design_status(self, qk):
        status = {"compiled": False, "run": False, "success": False, "assemblies": []}

        record = self.seqdb[qk]
        status["record"] = {
            "name": record.name,
            "length": len(record.seq),
            "id": record.id,
            "is_circular": is_circular(record),
        }

        if self.graphs.get(qk, None) is not None:
            status["compiled"] = True

        if self.results.get(qk, None) is not None:
            status["run"] = True
            for a in self.results[qk].assemblies:
                status["success"] = True
                summ_df = a.to_df()
                material = sum(list(summ_df["material"]))
                eff = functools.reduce(operator.mul, summ_df["efficiency"])

                comp = 0
                for c in summ_df["complexity"]:
                    if not c:
                        continue
                    elif c > comp:
                        comp = c

                status["assemblies"].append(
                    {
                        "cost": {
                            "material cost": round(material, 2),
                            "assembly efficiency": round(eff, 2),
                            "max synthesis complexity": round(comp, 2),
                        }
                    }
                )
        return status

    @property
    def metadata(self):
        return {
            "program": __title__,
            "version": __version__,
            "execution_trace": self._method_trace,
        }

    @property
    def status(self):
        """Return the statuses for each subdesign."""
        statuses = {}
        for qk in self.container_factory.containers():
            statuses[qk] = self._get_design_status(qk)
        return statuses

    @property
    def seqdb(self) -> Dict[str, SeqRecord]:
        return self._seqdb

    @property
    def results(self):
        return dict(self._results)

    @classmethod
    @functools.wraps(generate_fake_designs)
    def fake(
        cls,
        n_designs: int,
        span_cost: SpanCost = None,
        circular: bool = True,
        n_linear_seqs: int = 50,
        n_cyclic_seqs: int = 50,
        n_primers: int = 50,
        n_primers_from_templates: int = 50,
        shared_length: int = 0,
        cyclic_size_int: Tuple[int, int] = (3000, 10000),
        linear_size_int: Tuple[int, int] = (100, 4000),
        primer_size_int: Tuple[int, int] = (15, 60),
        plasmid_size_interval: Tuple[int, int] = (5000, 10000),
        chunk_size_interval: Tuple[int, int] = (100, 3000),
        random_chunk_prob_int: Tuple[float, float] = (0, 0.5),
        random_chunk_size_int: Tuple[int, int] = (100, 1000),
        return_with_library: bool = False,
        **kwargs,
    ):
        library = generate_fake_designs(
            n_designs=n_designs,
            circular=circular,
            n_linear_seqs=n_linear_seqs,
            n_cyclic_seqs=n_cyclic_seqs,
            n_primers=n_primers,
            n_primers_from_templates=n_primers_from_templates,
            cyclic_size_int=cyclic_size_int,
            linear_size_int=linear_size_int,
            primer_size_int=primer_size_int,
            plasmid_size_interval=plasmid_size_interval,
            chunk_size_interval=chunk_size_interval,
            random_chunk_prob_int=random_chunk_prob_int,
            random_chunk_size_int=random_chunk_size_int,
            design_sequence_similarity_length=shared_length,
        )
        designs = library["design"]
        plasmids = library["cyclic"]
        fragments = library["linear"]
        primers = library["short"]

        design = cls(span_cost=span_cost)
        design.add_materials(
            primers=primers, fragments=fragments, templates=plasmids, queries=designs
        )
        if return_with_library:
            return design, library
        return design

    def add_materials(
        self,
        primers: List[SeqRecord],
        templates: List[SeqRecord],
        queries: List[SeqRecord],
        fragments: List[SeqRecord] = None,
    ):
        if fragments is None:
            fragments = []
        self.add_primers(primers)
        fragments = self.filter_linear_records(fragments)
        self.add_templates(templates + fragments)
        self.add_queries(queries)
        self.add_fragments(fragments)

    def add_primers(self, primers: List[SeqRecord]):
        """Add primer sequences to materials list."""
        self.blast_factory.add_records(primers, self.PRIMERS)
        self.logger.info("Added {} primers".format(len(primers)))

    def add_templates(self, templates: List[SeqRecord]):
        """Add template sequences to materials list."""
        self.blast_factory.add_records(templates, self.TEMPLATES)
        self.logger.info("Added {} templates".format(len(templates)))

    def add_queries(self, queries: List[SeqRecord]):
        """Add goal/query sequences to materials list."""
        self.blast_factory.add_records(queries, self.QUERIES)
        self.logger.info("Added {} queries".format(len(queries)))

    def add_fragments(self, fragments: List[SeqRecord]):
        """Add fragment sequences to materials list."""
        self.blast_factory.add_records(fragments, self.FRAGMENTS)
        self.logger.info("Added {} queries".format(len(fragments)))
        # self.blast_factory.add_records(fragments, self.TEMPLATES)

    @classmethod
    def filter_linear_records(cls, records: List[SeqRecord]) -> List[SeqRecord]:
        """Return only linear records."""
        return [r for r in records if not is_circular(r)]

    @classmethod
    def filter_perfect_subject(cls, results: dict) -> List[dict]:
        """return only results whose subject is 100% aligned to query."""
        return [r for r in results if perfect_subject(r["subject"])]

    # non-threaded blast
    def _blast(self):
        # align templates
        template_blast = self.blast_factory(self.TEMPLATES, self.QUERIES)
        template_blast.blastn()
        template_results = template_blast.get_perfect()
        self.template_results = template_results
        self.logger.info("Number of template matches: {}".format(len(template_results)))
        self.container_factory.seqdb.update(template_blast.seq_db.records)

        # align fragments
        if self.blast_factory.record_groups[self.FRAGMENTS]:
            fragment_blast = self.blast_factory(self.FRAGMENTS, self.QUERIES)
            fragment_blast.blastn()
            fragment_results = self.filter_perfect_subject(fragment_blast.get_perfect())
            self.container_factory.seqdb.update(fragment_blast.seq_db.records)
            self.logger.info(
                "Number of perfect fragment matches: {}".format(len(fragment_results))
            )
        else:
            fragment_results = []
        self.fragment_results = fragment_results

        # align primers
        if self.blast_factory.record_groups[self.PRIMERS]:
            primer_blast = self.blast_factory(self.PRIMERS, self.QUERIES)
            primer_blast.blastn_short()
            primer_results = self.filter_perfect_subject(primer_blast.get_perfect())
            self.container_factory.seqdb.update(primer_blast.seq_db.records)
            self.logger.info(
                "Number of perfect primers: {}".format(len(primer_results))
            )
        else:
            primer_results = []
        self.primer_results = primer_results

        # initialize container factory
        query_keys = [rec.id for rec in self.blast_factory.record_groups[self.QUERIES]]
        query_keys = [self.blast_factory.db.get_origin_key(k) for k in query_keys]
        self.container_factory.initialize_empty(query_keys)

        # load results
        self.container_factory.load_blast_json(fragment_results, Constants.FRAGMENT)
        self.container_factory.load_blast_json(template_results, Constants.PCR_PRODUCT)
        self.container_factory.load_blast_json(primer_results, Constants.PRIMER)

    @property
    def containers(self) -> Dict[str, AlignmentContainer]:
        """Iterable of alignment containers in this design."""
        return self.container_factory.containers()

    @property
    def container_list(self) -> List[AlignmentContainer]:
        """List of alignment containers in this design."""
        return list(self.container_factory.containers().values())

    @property
    def query_keys(self) -> List[str]:
        """List of query keys in this design."""
        return list(self.container_factory.containers())

    def assemble_graphs(self, query_keys: Optional[List[str]] = None):
        """Assemble all assembly graphs for all queries in this design."""
        for query_key, container in self.logger.tqdm(
            self._filter_containers(query_keys).items(),
            "INFO",
            desc="assembling graphs",
        ):
            self.graphs[query_key], _ = assemble_graph(container, self.span_cost)

    def post_process_graphs(self, post_processor_kwargs: Optional[dict] = None):
        if post_processor_kwargs is None:
            post_processor_kwargs = {}
        for qk, graph in self.graphs.items():
            container = self.container_factory.containers()[qk]
            query = self.seqdb[qk]
            processor = AssemblyGraphPostProcessor(
                graph,
                query,
                self.span_cost,
                self.seqdb,
                container,
                **post_processor_kwargs,
            )
            processor.logger = self.logger(processor)
            processor()

    def _filter_containers(self, query_keys):
        if query_keys is None:
            return self.container_factory.containers()
        return {
            k: v
            for k, v in self.container_factory.containers().items()
            if k in query_keys
        }

    def uncompile(self):
        self.container_factory.reset()
        self._results = {}

    @abstractmethod
    def precompile(self):
        pass

    @abstractmethod
    def postcompile(self, post_process_kwargs):
        pass

    @log_metadata("compile", additional_metadata={"algorithm": ALGORITHM})
    def compile(self, post_process_kwargs: Dict = None):
        """Compile the materials list into assembly graphs."""
        self.precompile()

        self.logger.info("Assembling graphs")
        self.assemble_graphs()
        self.postcompile(post_process_kwargs)

    def _run(self, n_paths: int, post_processing_kwargs: Dict):
        """Run the design. Runs `compile` and `optimize`, returning results
        that can be accessed by `design.results` or by `design.out()`

        :param n_paths: max number of assemblies per design to design
        :return: results
        """
        self.compile(post_processing_kwargs)
        return self.optimize(n_paths=n_paths)

    def run(
        self,
        n_paths: int = None,
        post_processing_kwargs: Dict = None,
        n_jobs: int = None,
        job_size: int = None,
    ):
        if n_paths is None:
            n_paths = self.DEFAULT_N_ASSEMBLIES
        if n_jobs is None:
            n_jobs = self.DEFAULT_N_JOBS
        if job_size is None:
            job_size = self.DEFAULT_JOB_SIZE
        if n_jobs > 1:
            self._run_with_pool(
                n_jobs=n_jobs,
                job_size=job_size,
                post_processing_kwargs=post_processing_kwargs,
            )
        else:
            self._run(n_paths=n_paths, post_processing_kwargs=post_processing_kwargs)
        self._freeze_graphs()

    # TODO: better multithreaded loggers
    @staticmethod
    def _pooled_run(args: List[Tuple["Design", str]]) -> Dict[str, DesignResult]:
        design, qk, n_paths, post_processing_kwargs, thread = args
        design.logger.name = "THREAD {}: ".format(thread) + design.logger.name

        design.assemble_graphs(query_keys=qk)
        design.postcompile(post_processing_kwargs)
        design.optimize(n_paths=n_paths)
        return design.graphs, design.results

    def _run_with_pool(
        self,
        n_jobs: int = None,
        job_size: int = None,
        n_paths: int = None,
        post_processing_kwargs: dict = None,
    ):
        if n_paths is None:
            n_paths = self.DEFAULT_N_ASSEMBLIES
        if n_jobs is None:
            n_jobs = self.DEFAULT_N_JOBS
        if job_size is None:
            job_size = self.DEFAULT_JOB_SIZE
        self.precompile()
        query_keys = self.query_keys
        if not query_keys:
            raise DasiDesignException("There are no design sequences.")

        # divide query_keys
        chunks = list(chunkify(query_keys, job_size))
        print(chunks)
        args = list(
            zip(
                [self] * len(chunks),
                chunks,
                [n_paths] * len(chunks),
                [post_processing_kwargs] * len(chunks),
                range(len(chunks)),
            )
        )
        with Pool(processes=min(len(chunks), n_jobs)) as pool:
            graphs_and_results = pool.map(self._pooled_run, args)

        # update graphs and results
        combined_results = {}
        combined_graphs = {}
        for g, r in graphs_and_results:
            combined_results.update(r)
            combined_graphs.update(g)
        self.graphs = combined_graphs
        self._results = combined_results
        return self.results

    @property
    def is_compiled(self):
        """Return whether the design has been compiled.

        :return: True if compiled. False if otherwise.
        """
        if self.container_list:
            return True
        return False

    # def plot_matrix(self, matrix):
    # plot matrix
    # import pylab as plt
    # import seaborn as sns
    # import numpy as np
    #
    # plot_matrix = matrix.copy()
    # plot_matrix[plot_matrix == np.inf] = 10000
    # plot_matrix = np.nan_to_num(plot_matrix)
    #
    # fig = plt.figure(figsize=(24, 20))
    # ax = fig.gca()
    # step = 1
    # sns.heatmap(plot_matrix[::step, ::step], ax=ax)

    @staticmethod
    def _find_iter_alignment(a: int, b: int, alignments: Iterable[Alignment]):
        for align in alignments:
            if a == align.query_region.a and b == align.query_region.b:
                yield align

    @staticmethod
    def path_to_edge_costs(
        path: List[AssemblyNode], graph: nx.DiGraph
    ) -> List[Tuple[AssemblyNode, AssemblyNode, dict]]:
        arr = []
        for n1, n2 in pairwise(path):
            edata = graph[n1][n2]
            arr.append((n1, n2, edata))
        return arr

    @log_metadata("optimize", additional_metadata={"algorithm": ALGORITHM})
    def optimize(self, n_paths=DEFAULT_N_ASSEMBLIES) -> Dict[str, DesignResult]:

        if not self.container_list:
            raise DasiDesignException(
                "Design must be compiled before running optimization.'"
            )
        self._results = self._optimize(n_paths)
        return self._results

    def _freeze_graphs(self):
        for graph in self.graphs.values():
            nx.freeze(graph)

    def _optimize(self, n_paths) -> Dict[str, DesignResult]:
        """Finds the optimal paths for each query in the design."""
        results_dict = {}
        for query_key, graph, query_length, cyclic, result in self.logger.tqdm(
            self._collect_optimize_args(self.graphs),
            "INFO",
            desc="optimizing graphs (n_graphs={})".format(len(self.graphs)),
        ):

            container = self.containers[query_key]
            query = container.seqdb[query_key]
            cyclic = is_circular(query)
            results_dict[query_key] = result

            paths, costs = optimize_graph(graph, len(query), cyclic, n_paths)
            if not paths:
                query_rec = self.blast_factory.db.records[query_key]
                self.logger.error(
                    "\n\tThere were no solutions found for design '{}' ({}).\n\t"
                    "This sequence may be better synthesized. Use a tool such as JBEI's"
                    " BOOST.".format(query_rec.name, query_key)
                )
            result.add_assemblies(paths, ignore_invalid=True)
        return results_dict

    def _collect_optimize_args(
        self, graphs: Dict[str, nx.DiGraph]
    ) -> Tuple[str, nx.DiGraph, bool, dict]:
        for query_key, graph in self.logger.tqdm(
            graphs.items(), "INFO", desc="optimizing graphs"
        ):
            container = self.containers[query_key]
            query = container.seqdb[query_key]
            cyclic = is_circular(query)
            result = DesignResult(container=container, query_key=query_key, graph=graph)
            yield query_key, graph, len(query), cyclic, result

    # TODO: order keys
    # TODO: group identical reactions (identical output sequence)
    def to_df(self, assembly_index: int = 0) -> Tuple[pd.DataFrame, pd.DataFrame]:
        reaction_dfs = []
        summary_dfs = []
        design_json = {}
        for i, (qk, result) in enumerate(self.results.items()):
            if result.assemblies and result.assemblies[0]:
                assembly = result.assemblies[assembly_index]
                react_df = assembly.to_reaction_df()
                react_df["DESIGN_ID"] = i
                react_df["DESIGN_KEY"] = qk
                react_df["ASSEMBLY_ID"] = assembly_index
                reaction_dfs.append(react_df)

                summ_df = assembly.to_df()
                summ_df["DESIGN_ID"] = i
                summ_df["DESIGN_KEY"] = qk
                summ_df["ASSEMBLY_ID"] = assembly_index
                summary_dfs.append(summ_df)

                rec = self.seqdb[qk]
                design_json[qk] = {"NAME": rec.name, "FILE": None}
                if hasattr(rec, "from_file"):
                    design_json[qk]["FILE"] = rec.from_file
            else:
                msg = "Query {} {} yielded no assembly".format(self.seqdb[qk].name, i)
                self.logger.error(msg)

        colnames = [
            "DESIGN_ID",
            "DESIGN_KEY",
            "ASSEMBLY_ID",
            "REACTION_ID",
            "REACTION_NAME",
            "NAME",
            "TYPE",
            "KEY",
            "ROLE",
            "REGION",
            "SEQUENCE",
            "LENGTH",
            "META",
        ]
        if reaction_dfs:
            react_df = pd.concat(reaction_dfs)
        else:
            react_df = pd.DataFrame(columns=colnames)
        react_df.columns = colnames
        react_df.sort_values(
            by=["TYPE", "DESIGN_ID", "ASSEMBLY_ID", "REACTION_ID", "NAME", "ROLE"],
            inplace=True,
        )

        if summary_dfs:
            summ_df = pd.concat(summary_dfs)
        else:
            summ_df = pd.DataFrame()

        return react_df, summ_df, design_json

    def report(self):
        return Report(self)

    @property
    def last_run_start(self):
        x = self._method_run_times.get("optimize", None)
        if not x:
            return None
        else:
            return x[0]

    @property
    def last_run_end(self):
        x = self._method_run_times.get("optimize", None)
        if not x:
            return None
        else:
            return x[1]

    @property
    def last_compile_start(self):
        x = self._method_run_times.get("compile", None)
        if not x:
            return None
        else:
            return x[0]

    @property
    def last_compile_end(self):
        x = self._method_run_times.get("compile", None)
        if not x:
            return None
        else:
            return x[1]

    def out(
        self,
        fmt: str = "json",
        elim_extra_reactions: bool = False,
        query_keys: Optional[List[str]] = None,
    ):
        """Return the results of the design as a validates output JSON.

        The output JSON is follows the following schema, see
        :param fmt:
        :return:
        """
        if fmt.lower() == "json":
            output = dasi_design_to_output_json(
                self, elim_extra_reactions=elim_extra_reactions, query_keys=query_keys
            )
            validate_output(output)
            return output
        else:
            raise ValueError("Format '{}' not recognized".format(fmt))

    def out_jsonschema(self):
        """Return the output JSONschema."""
        return Schemas.output_schema



class Config:
    """Global DASi design configuration."""

    PRIMER3_MIN_ANNEAL_CHECK = 12  #: min n bases to check for misprimings
    PRIMER3_N_RUNS = (
        15  #: number of optimization steps to use for primer3 primer designs
    )
    ASSEMBLY_COST_THRESHOLD = 10000  #: maximum cost of an assembly

    ############
    # PARAMETERS
    ############
    PRIMER_MIN_BIND = 14  #: minimum primer binding for searching for primer alignments
    MIN_OVERLAP = 15  #: minimum overlap for searching for overlapping alignments
    MAX_HOMOLOGY = 100  #: maximum overlap for searching for overlapping alignments
    INF = 10.0 ** 6  #: almost infinity

    class SequenceScoringConfig:
        """Configuration for scoring sequences."""

        stats_repeat_window = 14  #: length of kmer to find sequence repeats
        stats_window = 20  #: length of window for sliding window calculations
        stats_hairpin_window = 20  #: length of kmer to find sequence hairpins
        mispriming_min_anneal = 12  #: minimum bp to look for misprimings
        mispriming_max_anneal = 30  #: maximum expected bp for a primer
        mispriming_penalty = 0.5  #: multiplier to apply to each mispriming
        complexity_threshold = (
            10
        )  #: threshold for sequence complexity that is not synthesizable by a vendor
        edge_threshold = 0.05  #: edges with efficiencies below this threshold will be
        # removed.
        not_synthesizable_efficiency = (
            0.1  #: efficiency value for sequence that is not synthesizable
        )
        pcr_length_range_efficiency_multiplier = [
            (4000, 5000, 0.8),
            (5000, 6000, 0.5),
            (6000, 1000, 0.2),
        ]  #: multiplier to apply to the efficiency if PCR is within the given length range
        partition_overlap = 30  #: overlap when partitioning a highly complex sequence
        partition_step_size = 10  #: step size to approximate optimal partition
        SCORE_COMPLEXITY = "SCORE_COMPLEXITY"
        SCORE_MISPRIMINGS = "SCORE_MISPRIMINGS"
        SCORE_LONG = "SCORE_LONG_PCR_PRODUCTS"
        PARTITION = "PARTITION_SYNTHETIC_SEQUENCES"
        post_process_stages = (
            SCORE_COMPLEXITY,
            SCORE_LONG,
            SCORE_MISPRIMINGS,
            PARTITION,
        )
